# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re

from .tokenizing import LineInfo


class TatSuException(Exception):
    pass


class ParseException(TatSuException):
    pass


class OptionSucceeded(ParseException):
    pass


class ParseError(ParseException):
    pass


class GrammarError(ParseError):
    pass


class CodegenError(ParseError):
    pass


class FailedSemantics(ParseException):
    pass


class FailedParse(ParseException):
    def __init__(self, lineinfo: LineInfo, stack: list[str], msg: str):
        # NOTE:
        #   Pass all arguments to super() to avoid pickling problems
        #       https://stackoverflow.com/questions/27993567/
        super().__init__(lineinfo, stack, msg)

        self.lineinfo = lineinfo
        self.stack = stack
        self.msg = msg
        self.pos = lineinfo.end

    @property
    def message(self):
        return self.msg

    def __str__(self):
        info = self.lineinfo
        template = '{}({}:{}) {} :\n{}\n{}^\n{}'
        text = info.text.rstrip()
        leading = re.sub(r'[^\t]', ' ', text)[: info.col]

        text = text.expandtabs()
        leading = leading.expandtabs()
        return template.format(
            info.filename,
            info.line + 1,
            info.col + 1,
            self.message.rstrip(),
            text,
            leading,
            '\n'.join(self.stack),
        )


class FailedToken(FailedParse):
    def __init__(self, tokenizer, stack, token):
        super().__init__(tokenizer, stack, token)
        self.token = token

    @property
    def message(self):
        return 'expecting {}'.format(repr(self.token).lstrip('u'))


class FailedPattern(FailedParse):
    pass


class FailedRef(FailedParse):
    def __init__(self, tokenizer, stack, name):
        super().__init__(tokenizer, stack, name)
        self.name = name

    @property
    def message(self):
        return f"Could not resolve reference to rule '{self.name}'"


class FailedChoice(FailedParse):
    @property
    def message(self):
        return 'No viable option'


class FailedLookahead(FailedParse):
    @property
    def message(self):
        return 'failed lookahead'


class FailedLeftRecursion(FailedParse):
    @property
    def message(self):
        return 'infinite left recursion'


class FailedExpectingEndOfText(FailedParse):
    pass


class KeywordError(FailedParse):
    pass


FailedKeywordSemantics = KeywordError



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# from: https://www.unicode.org/charts/nameslist/n_2500.html
from __future__ import annotations

from typing import NamedTuple

""" Box Drawing
    © 2003–2025 Unicode, Inc.
    Unicode and the Unicode Logo are registered trademarks of Unicode, Inc.,
    in the U.S. and other countries. For terms of use and license,
    see https://www.unicode.org/terms_of_use.html.
"""

# NOTE:
#   All of these characters are intended for compatibility
#   with character cell graphic sets in use prior to 1990.


class UC(NamedTuple):
    code: int
    ustr: str
    name: str


# fmt: off
# ruff: noqa: RUF001
BOX_DRAWING = {
    '🔚': UC(0x1F51A, '\U0001F51A', "END WITH LEFTWARDS ARROW ABOVE"),

    '≔': UC(0x2254, '\u2254', "COLON EQUALS"),
    '∅': UC(0x2205, '\u2205', "EMPTY SET"),
    '⚠': UC(0x26A0, '\u26A0', "WARNING SIGN"),
    '→': UC(0x2192, '\u2192', "RIGHTWARDS ARROW"),
    '↱': UC(0x21b1, '\u21b1', "UPWARDS ARROW WITH TIP RIGHTWARDS"),
    '✂': UC(0x2702, '\u2702', "BLACK SCISSORS"),
    '↦': UC(0x21A6, '\u21A6', "RIGHTWARDS ARROW FROM BAR"),
    '⇥': UC(0x21E5, '\u21E5', "RIGHTWARDS ARROW TO BAR"),
    '↠': UC(0x21A0, '\u21A0', "RIGHTWARDS TWO HEADED ARROW"),

    # Light and heavy solid lines
    '─': UC(0x2500, '\u2500', "BOX DRAWINGS LIGHT HORIZONTAL"),
    '━': UC(0x2501, '\u2501', "BOX DRAWINGS HEAVY HORIZONTAL"),
    '│': UC(0x2502, '\u2502', "BOX DRAWINGS LIGHT VERTICAL"),
    '┃': UC(0x2503, '\u2503', "BOX DRAWINGS HEAVY VERTICAL"),

    # Light and heavy dashed lines
    '┄': UC(0x2504, '\u2504', "BOX DRAWINGS LIGHT TRIPLE DASH HORIZONTAL"),
    '┅': UC(0x2505, '\u2505', "BOX DRAWINGS HEAVY TRIPLE DASH HORIZONTAL"),
    '┆': UC(0x2506, '\u2506', "BOX DRAWINGS LIGHT TRIPLE DASH VERTICAL"),
    '┇': UC(0x2507, '\u2507', "BOX DRAWINGS HEAVY TRIPLE DASH VERTICAL"),
    '┈': UC(0x2508, '\u2508', "BOX DRAWINGS LIGHT QUADRUPLE DASH HORIZONTAL"),
    '┉': UC(0x2509, '\u2509', "BOX DRAWINGS HEAVY QUADRUPLE DASH HORIZONTAL"),
    '┊': UC(0x250A, '\u250A', "BOX DRAWINGS LIGHT QUADRUPLE DASH VERTICAL"),
    '┋': UC(0x250B, '\u250B', "BOX DRAWINGS HEAVY QUADRUPLE DASH VERTICAL"),

    # Light and heavy line box components
    '┌': UC(0x250C, '\u250C', "BOX DRAWINGS LIGHT DOWN AND RIGHT"),
    '┍': UC(0x250D, '\u250D', "BOX DRAWINGS DOWN LIGHT AND RIGHT HEAVY"),
    '┎': UC(0x250E, '\u250E', "BOX DRAWINGS DOWN HEAVY AND RIGHT LIGHT"),
    '┏': UC(0x250F, '\u250F', "BOX DRAWINGS HEAVY DOWN AND RIGHT"),
    '┐': UC(0x2510, '\u2510', "BOX DRAWINGS LIGHT DOWN AND LEFT"),
    '┑': UC(0x2511, '\u2511', "BOX DRAWINGS DOWN LIGHT AND LEFT HEAVY"),
    '┒': UC(0x2512, '\u2512', "BOX DRAWINGS DOWN HEAVY AND LEFT LIGHT"),
    '┓': UC(0x2513, '\u2513', "BOX DRAWINGS HEAVY DOWN AND LEFT"),
    '└': UC(0x2514, '\u2514', "BOX DRAWINGS LIGHT UP AND RIGHT"),
    '┕': UC(0x2515, '\u2515', "BOX DRAWINGS UP LIGHT AND RIGHT HEAVY"),
    '┖': UC(0x2516, '\u2516', "BOX DRAWINGS UP HEAVY AND RIGHT LIGHT"),
    '┗': UC(0x2517, '\u2517', "BOX DRAWINGS HEAVY UP AND RIGHT"),
    '┘': UC(0x2518, '\u2518', "BOX DRAWINGS LIGHT UP AND LEFT"),
    '┙': UC(0x2519, '\u2519', "BOX DRAWINGS UP LIGHT AND LEFT HEAVY"),
    '┚': UC(0x251A, '\u251A', "BOX DRAWINGS UP HEAVY AND LEFT LIGHT"),
    '┛': UC(0x251B, '\u251B', "BOX DRAWINGS HEAVY UP AND LEFT"),
    '├': UC(0x251C, '\u251C', "BOX DRAWINGS LIGHT VERTICAL AND RIGHT"),
    '┝': UC(0x251D, '\u251D', "BOX DRAWINGS VERTICAL LIGHT AND RIGHT HEAVY"),
    '┞': UC(0x251E, '\u251E', "BOX DRAWINGS UP HEAVY AND RIGHT DOWN LIGHT"),
    '┟': UC(0x251F, '\u251F', "BOX DRAWINGS DOWN HEAVY AND RIGHT UP LIGHT"),
    '┠': UC(0x2520, '\u2520', "BOX DRAWINGS VERTICAL HEAVY AND RIGHT LIGHT"),
    '┡': UC(0x2521, '\u2521', "BOX DRAWINGS DOWN LIGHT AND RIGHT UP HEAVY"),
    '┢': UC(0x2522, '\u2522', "BOX DRAWINGS UP LIGHT AND RIGHT DOWN HEAVY"),
    '┣': UC(0x2523, '\u2523', "BOX DRAWINGS HEAVY VERTICAL AND RIGHT"),
    '┤': UC(0x2524, '\u2524', "BOX DRAWINGS LIGHT VERTICAL AND LEFT"),
    '┥': UC(0x2525, '\u2525', "BOX DRAWINGS VERTICAL LIGHT AND LEFT HEAVY"),
    '┦': UC(0x2526, '\u2526', "BOX DRAWINGS UP HEAVY AND LEFT DOWN LIGHT"),
    '┧': UC(0x2527, '\u2527', "BOX DRAWINGS DOWN HEAVY AND LEFT UP LIGHT"),
    '┨': UC(0x2528, '\u2528', "BOX DRAWINGS VERTICAL HEAVY AND LEFT LIGHT"),
    '┩': UC(0x2529, '\u2529', "BOX DRAWINGS DOWN LIGHT AND LEFT UP HEAVY"),
    '┪': UC(0x252A, '\u252A', "BOX DRAWINGS UP LIGHT AND LEFT DOWN HEAVY"),
    '┫': UC(0x252B, '\u252B', "BOX DRAWINGS HEAVY VERTICAL AND LEFT"),
    '┬': UC(0x252C, '\u252C', "BOX DRAWINGS LIGHT DOWN AND HORIZONTAL"),
    '┭': UC(0x252D, '\u252D', "BOX DRAWINGS LEFT HEAVY AND RIGHT DOWN LIGHT"),
    '┮': UC(0x252E, '\u252E', "BOX DRAWINGS RIGHT HEAVY AND LEFT DOWN LIGHT"),
    '┯': UC(0x252F, '\u252F', "BOX DRAWINGS DOWN LIGHT AND HORIZONTAL HEAVY"),
    '┰': UC(0x2530, '\u2530', "BOX DRAWINGS DOWN HEAVY AND HORIZONTAL LIGHT"),
    '┱': UC(0x2531, '\u2531', "BOX DRAWINGS RIGHT LIGHT AND LEFT DOWN HEAVY"),
    '┲': UC(0x2532, '\u2532', "BOX DRAWINGS LEFT LIGHT AND RIGHT DOWN HEAVY"),
    '┳': UC(0x2533, '\u2533', "BOX DRAWINGS HEAVY DOWN AND HORIZONTAL"),
    '┴': UC(0x2534, '\u2534', "BOX DRAWINGS LIGHT UP AND HORIZONTAL"),
    '┵': UC(0x2535, '\u2535', "BOX DRAWINGS LEFT HEAVY AND RIGHT UP LIGHT"),
    '┶': UC(0x2536, '\u2536', "BOX DRAWINGS RIGHT HEAVY AND LEFT UP LIGHT"),
    '┷': UC(0x2537, '\u2537', "BOX DRAWINGS UP LIGHT AND HORIZONTAL HEAVY"),
    '┸': UC(0x2538, '\u2538', "BOX DRAWINGS UP HEAVY AND HORIZONTAL LIGHT"),
    '┹': UC(0x2539, '\u2539', "BOX DRAWINGS RIGHT LIGHT AND LEFT UP HEAVY"),
    '┺': UC(0x253A, '\u253A', "BOX DRAWINGS LEFT LIGHT AND RIGHT UP HEAVY"),
    '┻': UC(0x253B, '\u253B', "BOX DRAWINGS HEAVY UP AND HORIZONTAL"),
    '┼': UC(0x253C, '\u253C', "BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL"),
    '┽': UC(0x253D, '\u253D', "BOX DRAWINGS LEFT HEAVY AND RIGHT VERTICAL LIGHT"),
    '┾': UC(0x253E, '\u253E', "BOX DRAWINGS RIGHT HEAVY AND LEFT VERTICAL LIGHT"),
    '┿': UC(0x253F, '\u253F', "BOX DRAWINGS VERTICAL LIGHT AND HORIZONTAL HEAVY"),
    '╀': UC(0x2540, '\u2540', "BOX DRAWINGS UP HEAVY AND DOWN HORIZONTAL LIGHT"),
    '╁': UC(0x2541, '\u2541', "BOX DRAWINGS DOWN HEAVY AND UP HORIZONTAL LIGHT"),
    '╂': UC(0x2542, '\u2542', "BOX DRAWINGS VERTICAL HEAVY AND HORIZONTAL LIGHT"),
    '╃': UC(0x2543, '\u2543', "BOX DRAWINGS LEFT UP HEAVY AND RIGHT DOWN LIGHT"),
    '╄': UC(0x2544, '\u2544', "BOX DRAWINGS RIGHT UP HEAVY AND LEFT DOWN LIGHT"),
    '╅': UC(0x2545, '\u2545', "BOX DRAWINGS LEFT DOWN HEAVY AND RIGHT UP LIGHT"),
    '╆': UC(0x2546, '\u2546', "BOX DRAWINGS RIGHT DOWN HEAVY AND LEFT UP LIGHT"),
    '╇': UC(0x2547, '\u2547', "BOX DRAWINGS DOWN LIGHT AND UP HORIZONTAL HEAVY"),
    '╈': UC(0x2548, '\u2548', "BOX DRAWINGS UP LIGHT AND DOWN HORIZONTAL HEAVY"),
    '╉': UC(0x2549, '\u2549', "BOX DRAWINGS RIGHT LIGHT AND LEFT VERTICAL HEAVY"),
    '╊': UC(0x254A, '\u254A', "BOX DRAWINGS LEFT LIGHT AND RIGHT VERTICAL HEAVY"),
    '╋': UC(0x254B, '\u254B', "BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL"),

    # Light and heavy dashed lines
    '╌': UC(0x254C, '\u254C', "BOX DRAWINGS LIGHT DOUBLE DASH HORIZONTAL"),
    '╍': UC(0x254D, '\u254D', "BOX DRAWINGS HEAVY DOUBLE DASH HORIZONTAL"),
    '╎': UC(0x254E, '\u254E', "BOX DRAWINGS LIGHT DOUBLE DASH VERTICAL"),
    '╏': UC(0x254F, '\u254F', "BOX DRAWINGS HEAVY DOUBLE DASH VERTICAL"),

    # Double lines
    '═': UC(0x2550, '\u2550', "BOX DRAWINGS DOUBLE HORIZONTAL"),
    '║': UC(0x2551, '\u2551', "BOX DRAWINGS DOUBLE VERTICAL"),

    # Light and double line box components
    '╒': UC(0x2552, '\u2552', "BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE"),
    '╓': UC(0x2553, '\u2553', "BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE"),
    '╔': UC(0x2554, '\u2554', "BOX DRAWINGS DOUBLE DOWN AND RIGHT"),
    '╕': UC(0x2555, '\u2555', "BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE"),
    '╖': UC(0x2556, '\u2556', "BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE"),
    '╗': UC(0x2557, '\u2557', "BOX DRAWINGS DOUBLE DOWN AND LEFT"),
    '╘': UC(0x2558, '\u2558', "BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE"),
    '╙': UC(0x2559, '\u2559', "BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE"),
    '╚': UC(0x255A, '\u255A', "BOX DRAWINGS DOUBLE UP AND RIGHT"),
    '╛': UC(0x255B, '\u255B', "BOX DRAWINGS UP SINGLE AND LEFT DOUBLE"),
    '╜': UC(0x255C, '\u255C', "BOX DRAWINGS UP DOUBLE AND LEFT SINGLE"),
    '╝': UC(0x255D, '\u255D', "BOX DRAWINGS DOUBLE UP AND LEFT"),
    '╞': UC(0x255E, '\u255E', "BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE"),
    '╟': UC(0x255F, '\u255F', "BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE"),
    '╠': UC(0x2560, '\u2560', "BOX DRAWINGS DOUBLE VERTICAL AND RIGHT"),
    '╡': UC(0x2561, '\u2561', "BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE"),
    '╢': UC(0x2562, '\u2562', "BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE"),
    '╣': UC(0x2563, '\u2563', "BOX DRAWINGS DOUBLE VERTICAL AND LEFT"),
    '╤': UC(0x2564, '\u2564', "BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE"),
    '╥': UC(0x2565, '\u2565', "BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE"),
    '╦': UC(0x2566, '\u2566', "BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL"),
    '╧': UC(0x2567, '\u2567', "BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE"),
    '╨': UC(0x2568, '\u2568', "BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE"),
    '╩': UC(0x2569, '\u2569', "BOX DRAWINGS DOUBLE UP AND HORIZONTAL"),
    '╪': UC(0x256A, '\u256A', "BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE"),
    '╫': UC(0x256B, '\u256B', "BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE"),
    '╬': UC(0x256C, '\u256C', "BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL"),

    # Character cell arcs
    '╭': UC(0x256D, '\u256D', "BOX DRAWINGS LIGHT ARC DOWN AND RIGHT"),
    '╮': UC(0x256E, '\u256E', "BOX DRAWINGS LIGHT ARC DOWN AND LEFT"),
    '╯': UC(0x256F, '\u256F', "BOX DRAWINGS LIGHT ARC UP AND LEFT"),
    '╰': UC(0x2570, '\u2570', "BOX DRAWINGS LIGHT ARC UP AND RIGHT"),

    # Character cell diagonals
    # For a more extensive set of legacy terminal graphic character cell diagonals,
    # see also 1FBA0-1FBAE in the Symbols for Legacy Computing block.
    '╱': UC(0x2571, '\u2571', "BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT"),
    '╲': UC(0x2572, '\u2572', "BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT"),
    '╳': UC(0x2573, '\u2573', "BOX DRAWINGS LIGHT DIAGONAL CROSS"),

    # Light and heavy half lines
    '╴': UC(0x2574, '\u2574', "BOX DRAWINGS LIGHT LEFT"),
    '╵': UC(0x2575, '\u2575', "BOX DRAWINGS LIGHT UP"),
    '╶': UC(0x2576, '\u2576', "BOX DRAWINGS LIGHT RIGHT"),
    '╷': UC(0x2577, '\u2577', "BOX DRAWINGS LIGHT DOWN"),
    '╸': UC(0x2578, '\u2578', "BOX DRAWINGS HEAVY LEFT"),
    '╹': UC(0x2579, '\u2579', "BOX DRAWINGS HEAVY UP"),
    '╺': UC(0x257A, '\u257A', "BOX DRAWINGS HEAVY RIGHT"),
    '╻': UC(0x257B, '\u257B', "BOX DRAWINGS HEAVY DOWN"),

    # Mixed light and heavy lines
    '╼': UC(0x257C, '\u257C', "BOX DRAWINGS LIGHT LEFT AND HEAVY RIGHT"),
    '╽': UC(0x257D, '\u257D', "BOX DRAWINGS LIGHT UP AND HEAVY DOWN"),
    '╾': UC(0x257E, '\u257E', "BOX DRAWINGS HEAVY LEFT AND LIGHT RIGHT"),
    '╿': UC(0x257F, '\u257F', "BOX DRAWINGS HEAVY UP AND LIGHT DOWN"),
}



from __future__ import annotations

from tatsu.tool import compile
from tatsu.util import asjsons  # noqa: F401


def test_alert_interpolation(trace=True):
    input = '42 69'
    grammar = r"""
            start = a:number b: number i:^`"seen: {a}, {b}"` $ ;
            number::int = /\d+/ ;
    """
    model = compile(grammar)
    # print(asjsons(model))
    ast = model.parse(input, trace=trace)
    assert ast == {'a': '42', 'b': '69', 'i': 'seen: 42, 69'}



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
from itertools import chain
from typing import Any

from .. import grammars as model
from ..ast import AST
from ..grammars import Model


def camel2py(name: Any) -> str:
    return re.sub(
        r'([a-z0-9])([A-Z])',
        lambda m: m.group(1) + '_' + m.group(2).lower(),
        str(name),
    )


class ANTLRSemantics:
    def __init__(self, name):
        self.name = name
        self.tokens = {}
        self.token_rules = {}
        self.synthetic_rules = []

    def grammar(self, ast: AST) -> model.Grammar:
        return model.Grammar(
            self.name,
            [r for r in chain(ast.rules, self.synthetic_rules) if r is not None],
        )

    def rule(self, ast: AST) -> model.Rule | None:
        name = camel2py(ast.name)
        exp = ast.exp
        if name[0].isupper():
            name = name.upper()
            if isinstance(exp, model.Token):
                if name in self.token_rules:
                    self.token_rules[name].exp = exp  # it is a model.Decorator?
                else:
                    self.token_rules[name] = exp
                return None
            elif not ast.fragment and not isinstance(exp, model.Model):
                ref = model.Call(name.lower())
                if name in self.token_rules:
                    self.token_rules[name].exp = ref
                else:
                    self.token_rules[name] = ref
                name = name.lower()

        return model.Rule(ast=ast, name=name, params=ast.params, kwparams=ast.kwparams)

    def alternatives(self, ast: AST) -> model.Model:
        options = [o for o in ast.options if o is not None]
        if len(options) == 1:
            return options[0]
        else:
            options = [model.Option(o) for o in options]
            return model.Choice(options)

    def elements(self, ast: list[model.Model]) -> model.Model:
        elements = [e for e in ast if e is not None]
        if not elements:
            return model.Void()
        elif len(elements) == 1:
            return elements[0]
        else:
            return model.Sequence(elements)

    def predicate_or_action(self, ast: Any) -> None:
        return None

    def named(self, ast: AST) -> model.Named:
        if ast.force_list:
            return model.NamedList(ast)
        else:
            return model.Named(ast)

    def syntactic_predicate(self, ast: Any) -> None:
        return None

    def optional(self, ast: Model) -> model.Optional:
        if isinstance(ast, model.Group | model.Optional):
            ast = ast.exp
        return model.Optional(ast)

    def closure(self, ast: Model) -> model.Closure:
        if isinstance(ast, model.Group | model.Optional):
            ast = ast.exp
        return model.Closure(ast)

    def positive_closure(self, ast: Model) -> model.Closure:
        if isinstance(ast, model.Group):
            ast = ast.exp
        return model.PositiveClosure(ast)

    def negative(self, ast: Model) -> model.Sequence:
        neg = model.NegativeLookahead(ast)
        any = model.Pattern('.')
        return model.Sequence([neg, any])

    def subexp(self, ast: Model) -> model.Group:
        return model.Group(ast)

    def regexp(self, ast: AST) -> model.Pattern:
        pattern = ''.join(ast)
        regex = re.compile(pattern)
        return model.Pattern(regex.pattern)

    def charset_optional(self, ast: AST) -> str:
        return f'{ast}?'

    def charset_closure(self, ast: AST) -> str:
        return f'{ast}*'

    def charset_positive_closure(self, ast: AST) -> str:
        return f'{ast}+'

    def charset_or(self, ast: AST) -> str:
        return '[{}]'.format(''.join(ast))

    def charset_negative_or(self, ast: AST) -> str:
        return '[^{}]'.format(''.join(ast))

    @staticmethod
    def escape(s: str) -> str:
        return ''.join('\\' + c if c in '[]().*+{}^$' else c for c in s)

    def charset_atom(self, ast: AST) -> AST:
        return ast

    def charset_char(self, ast: AST) -> AST:
        return ast

    def charset_range(self, ast: AST) -> str:
        return f'{ast.first}-{ast.last}'

    def newranges(self, ast: AST) -> model.Pattern:
        pattern = ''.join(ast)
        regex = re.compile(pattern)
        return model.Pattern(regex.pattern)

    def newrange(self, ast: AST) -> str:
        pattern = '[{}]{}'.format(ast.range, ast.repeat or '')
        re.compile(pattern)
        return pattern

    def negative_newrange(self, ast: AST) -> str:
        pattern = '[^{}]{}'.format(ast.range, ast.repeat or '')
        re.compile(pattern)
        return pattern

    def call(self, ast: str) -> model.Call:
        assert ast[0].islower()
        return model.Call(camel2py(ast))

    def any(self, ast: AST) -> model.Pattern:
        return model.Pattern(r'\w+|\S+')

    def string(self, ast: AST) -> model.Token:
        text = str(ast)
        if isinstance(text, list):
            text = ''.join(text)
        return model.Token(text)

    def eof(self, ast: AST) -> model.EOF:
        return model.EOF()

    def token(self, ast: AST) -> model.Token | model.Void:
        name = ast.name
        if ast.exp:
            exp = model.Token(ast.value)
            self.tokens[name] = exp
        else:
            exp = model.Void()
            rule = model.Rule(ast=exp, name=name)
            self.synthetic_rules.append(rule)
        return exp

    def token_ref(self, ast: AST) -> model.Model:
        name = camel2py(ast).upper()

        value = self.tokens.get(name)
        if value and isinstance(value, model.Model):
            return value

        if name in self.token_rules:
            exp = self.token_rules[name]
        else:
            exp = model.Decorator(model.Call(name))
            self.token_rules[name] = exp
        return exp



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# CAVEAT: THIS LEGACY MODULE IS KEPT ONLY FOR BACKWARDS COMPATIBILITY
from __future__ import annotations

from ..exceptions import CodegenError
from .cgbase import CodeGenerator, ModelRenderer
from .genmodel import modelgen
from .genpython import codegen

__all__ = ['CodeGenerator', 'ModelRenderer', 'CodegenError', 'codegen', 'modelgen']



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import io
from collections.abc import Iterator
from contextlib import contextmanager

from ..util import trim

BLACK_LINE_LENGTH = 88


class IndentPrintMixin:

    def __init__(self, indent_amount: int = 4):
        self.indent_amount = indent_amount
        self.indent_stack: list[int] = [0]
        self.output_stream = io.StringIO()

    def printed_text(self):
        return self.output_stream.getvalue()

    def print(self, *args, **kwargs):
        args = [trim(str(arg)) for arg in args if arg is not None]
        lines = self.as_printed_lines(*args, **kwargs)
        self._do_print_lines(lines, **kwargs)

    def as_printed(self, *args, **kwargs) -> str:
        lines = self.as_printed_lines(*args, **kwargs)
        return '\n'.join(lines)

    def as_printed_lines(self, *args, **kwargs) -> list[str]:
        text = self.io_print(*args, **kwargs)
        return self.indented_lines(text)

    def fitsfmt(self, line: str, addindents: int = 0):
        assert addindents >= 0
        total = self.indentation + len(line)
        total += addindents * self.indent_amount
        return total <= BLACK_LINE_LENGTH

    @contextmanager
    def indent(self, amount: int | None = None) -> Iterator:
        assert amount is None or amount >= 0
        amount = amount if amount is not None else self.indent_amount

        self.indent_stack.append(amount + self.indent_stack[-1])
        try:
            yield
        finally:
            self.indent_stack.pop()

    @property
    def indentation(self) -> int:
        return self.indent_stack[-1]

    @property
    def indentstr(self) -> str:
        return ' ' * self.indentation

    @staticmethod
    def io_print(*args, **kwargs) -> str:
        kwargs.pop('file', None)  # do not allow redirection of output
        with io.StringIO() as output:
            print(*args, file=output, **kwargs)
            return output.getvalue()

    def _do_print_lines(self, lines: list[str] | None = None, **kwargs):
        if not lines:
            print(file=self.output_stream, **kwargs)
            return

        for line in lines:
            print(line, file=self.output_stream, **kwargs)

    def indented_lines(self, text: str) -> list[str]:
        text = trim(text)
        return [
            (self.indentstr + line).rstrip() for line in text.splitlines(keepends=False)
        ]



#!/usr/bin/env python3
#
# \u7ADC TatSu documentation build configuration file, created by
# sphinx-quickstart on Mon May  1 18:01:31 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('.'))
del sys
del os
import tatsu

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',

    "myst_parser",
    "sphinx_copybutton",  # Recommended for "Copy" icons on code blocks
]

# MyST-Parser Settings
myst_enable_extensions = [
    "colon_fence",     # Allows using ::: for notes/admonitions
    "deflist",         # Allows definition lists (Term : Definition)
    "dollarmath",      # Allows LaTeX math via $$ ... $$
    "fieldlist",       # Allows GitHub-style metadata fields
    "tasklist",        # Allows - [ ] checkbox lists
    "html_image",      # Allows using standard <img src="..."> tags
]

# Auto-generate anchors for headers (so you can link to #my-header)
myst_heading_anchors = 4  # Generates anchors for H1, H2, H3, ...

copybutton_prompt_text = r">>> |\$ "
copybutton_only_copy_prompt_lines = True
copybutton_use_regexp = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
# source_suffix = '.rst'
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = '\u7ADC TatSu'
copyright = '2017-2026 Juancarlo Añez'
author = 'Juancarlo Añez'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = tatsu.__version__
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = 'sphinx_rtd_theme'
# html_theme = 'alabaster'
html_theme = "pydata_sphinx_theme"

html_context = {
   "default_mode": "auto"  # Options: "light", "dark", or "auto"
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}
html_theme_options = {
    # Controls how deep the sidebar navigation goes
    'navigation_depth': 5,

    # If True, the sidebar will only show document titles, not subheadings
    # 'titles_only': False,

    # If True, headings will collapse as you navigate
    'collapse_navigation': True,

    "logo": {
        "text": "TatSu",
        "image_light": "_static/tatsu_logo.svg",
        "image_dark": "_static/tatsu_logo.svg",
    },
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/neogeny/TatSu",
            "icon": "fa-brands fa-github",
        },
    ],
    "header_links_before_dropdown": 4,  # How many nav links to show before "More"
    "show_prev_next": True,             # Previous/Next buttons at the bottom
    "navbar_align": "content",          # Align the top nav with the text content
}

# Add any paths that contain custom themes here, relative to this directory.
# Add path to the RTD explicitly to robustify builds (otherwise might
# fail in a clean Debian build env)
html_theme_path = []

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static', '_static/css']


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'tatsudoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'tatsu.tex', '\u7ADC TatSu Documentation',
     'Juancarlo Añez', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, '\u7ADC TatSu', '\u7ADC TatSu Documentation',
     [author], 1),
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, '\u7ADC TatSu', '\u7ADC TatSu Documentation',
     author, '\u7ADC TatSu', 'One line description of project.',
     'Miscellaneous'),
]


html_sidebars = {
    '**': [
        'globaltoc.html',
        'relations.html',
        'sourcelink.html',
        'searchbox.html',
    ],
}


def setup(app):
    app.add_css_file('css/custom.css')  # may also be an URL



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys

import pytest

from tatsu import synth
from tatsu.builder import BuilderConfig, ModelBuilderSemantics, TypeResolutionError
from tatsu.exceptions import FailedParse, FailedToken
from tatsu.objectmodel import Node
from tatsu.tool import compile, parse


class MyNode:
    def __init__(self, ast):
        pass


class BType:
    pass


class AType(BType):
    pass


def test_semantics_not_class():
    grammar = r"""
        start::sum = {number}+ $ ;
        number::int = /\d+/ ;
    """
    text = '5 4 3 2 1'
    bad_semantics = ModelBuilderSemantics  # NOTE: the class
    semantics = ModelBuilderSemantics()  # NOTE: the class

    with pytest.raises(
        TypeError,
        match=r'semantics must be an object instance or None.*',
    ):
        compile(grammar, semantics=bad_semantics)
        compile(grammar, semantics=bad_semantics)

    model = compile(grammar, 'test')
    with pytest.raises(
        TypeError,
        match=r'semantics must be an object instance or None.*',
    ):
        model.parse(text, semantics=bad_semantics)

    ast = model.parse(text, semantics=semantics)
    assert ast == 15


def test_builder_semantics():
    grammar = r"""
        start::sum = {number}+ $ ;
        number::int = /\d+/ ;
    """
    text = '5 4 3 2 1'

    semantics = ModelBuilderSemantics()
    model = compile(grammar, 'test')
    ast = model.parse(text, semantics=semantics)
    assert ast == 15

    import functools

    dotted = functools.partial(str.join, '.')
    # WARNING: this defeats all TatSu knows about functions and types
    #  dotted.__name__ = 'dotted'

    grammar = r"""
        start::dotted = {number}+ $ ;
        number = /\d+/ ;
    """

    semantics = ModelBuilderSemantics(constructors=[dotted])
    model = compile(grammar, 'test')
    with pytest.raises(
        TypeResolutionError,
        match=r"Could not find constructor for type 'dotted'",
    ):
        ast = model.parse(text, semantics=semantics)
        assert ast == '5.4.3.2.1'


def test_builder_subclassing():
    registry = getattr(synth, '__registry')

    grammar = """
        @@grammar :: Test
        start::A::B::C = $ ;
    """

    model = compile(grammar, asmodel=True)
    model.parse('')

    print(f'{registry=}')
    A = registry['A']
    B = registry['B']
    C = registry['C']

    assert issubclass(A, B) and issubclass(A, synth.SynthNode) and issubclass(A, Node)
    assert issubclass(B, C) and issubclass(B, synth.SynthNode) and issubclass(A, Node)
    assert issubclass(C, synth.SynthNode) and issubclass(C, Node)


def test_builder_basetype_codegen():
    grammar = """
        @@grammar :: Test
        start::A::B::C = a:() b:() $ ;
        second::D::A = ();
        third = ();
    """

    from tatsu.tool import to_python_model

    src = to_python_model(grammar, basetype=MyNode)
    # print(src[:1000])

    globals = {}
    exec(src, globals)  # pylint: disable=W0122
    semantics = globals['TestModelBuilderSemantics']()

    A = globals['A']
    B = globals['B']
    C = globals['C']
    D = globals['D']

    model = compile(grammar, semantics=semantics)
    ast = model.parse('', semantics=semantics)
    # print(f'AST({type(ast)}=', ast)

    assert isinstance(ast, MyNode), A.__mro__
    assert isinstance(ast, (A, B, C))
    assert hasattr(ast, 'a')
    assert hasattr(ast, 'b')

    assert issubclass(D, A | B | C)


def test_optional_attributes():
    grammar = r"""
        foo::Foo = left:identifier [ ':' right:identifier ] $ ;
        identifier = /\w+/ ;
    """

    grammar = compile(grammar)

    a = grammar.parse('foo : bar', semantics=ModelBuilderSemantics())
    assert a.left == 'foo'
    assert a.right == 'bar'

    b = grammar.parse('foo', semantics=ModelBuilderSemantics())
    assert b.left == 'foo'
    assert b.right is None


def test_constant_math():
    grammar = r"""
        start = a:`7` b:`2` @:```{a} / {b}``` $ ;
    """
    result = parse(grammar, '', trace=True)
    assert result <= 3.5 <= result


def test_constant_deep_eval():
    grammar = r"""
        start =
            a:A b:B
            @:```{a} / {b}```
            $
        ;

        A = number @:`7` ;
        B = number @:`0` ;
        number::int = /\d+/ ;
    """
    model = compile(grammar)

    with pytest.raises(
        FailedParse,
        match=r'Error evaluating constant.*ZeroDivisionError',
    ):
        # NOTE: only with multiple evaluation passes on constants
        model.parse('42 84', trace=True)


def test_builder_types():
    grammar = """
        @@grammar :: Test
        start::AType::BType = $ ;
    """

    builderconfig = BuilderConfig(basetype=BType, constructors=[AType, BType])
    ast = parse(grammar, '', builderconfig=builderconfig)
    assert type(ast) is AType
    assert isinstance(ast, BType)
    assert not isinstance(ast, synth.SynthNode)


def test_builder_nodedefs():
    grammar = """
        @@grammar :: Test
        start::AType::BType = $ ;
    """

    thismodule = sys.modules[__name__]
    builderconfig = BuilderConfig(typedefs=[thismodule], synthok=False)
    ast = parse(grammar, '', builderconfig=builderconfig)
    assert type(ast) is AType
    assert isinstance(ast, BType)
    assert not isinstance(ast, synth.SynthNode)


def test_ast_per_option():
    grammar = """
        start = options $ ;

        options =
            | a:'a' [b:'b']
            | c:'c' [d:'d']
            ;
    """

    # NOTE:
    #   Prove each option in a choide has its own version of the AST

    ast = parse(grammar, 'a b')
    assert ast == {'a': 'a', 'b': 'b'}
    assert 'c' not in ast
    assert 'd' not in ast

    ast = parse(grammar, 'c d')
    assert ast == {'c': 'c', 'd': 'd'}
    assert 'a' not in ast
    assert 'b' not in ast

    ast = parse(grammar, 'a')
    assert ast == {'a': 'a', 'b': None}
    assert 'b' in ast
    assert 'c' not in ast
    assert 'd' not in ast

    ast = parse(grammar, 'c')
    assert ast == {'c': 'c', 'd': None}
    assert 'd' in ast
    assert 'a' not in ast
    assert 'b' not in ast


def test_ast_names_accumulate():
    grammar = """
        start = options $ ;

        options =
            | a:'a' ([b:'b'] {x:'x'})
            | c:'c' ([d:'d'] y:{'y'})
            ;
    """

    # NOTE:
    #   Prove named elements accumulat
    ast = parse(grammar, 'a')
    assert ast == {'a': 'a', 'b': None}

    ast = parse(grammar, 'a x')
    assert ast == {'a': 'a', 'b': None, 'x': 'x'}

    ast = parse(grammar, 'a x x')
    assert ast == {'a': 'a', 'b': None, 'x': ['x', 'x']}

    # NOTE:
    #   Prove naming closures always closure
    ast = parse(grammar, 'c')
    assert ast == {'c': 'c', 'd': None, 'y': []}

    ast = parse(grammar, 'c y')
    assert ast == {'c': 'c', 'd': None, 'y': ['y']}

    ast = parse(grammar, 'c y y')
    assert ast == {'c': 'c', 'd': None, 'y': ['y', 'y']}


def test_cut_scope():
    grammar = """
        start = failcut | failchoice | succeed $ ;

        failcut = 'a' ~ 'y' ;

        failchoice =
            | 'a' ~ 'b'
            | 'a' 'c' 'd'
            ;

        succeed = ('a' ~ 'y' | 'b' 'z') | 'a' 'x' ;
    """

    ast = parse(grammar, 'a x')
    assert ast == ('a', 'x')

    ast = parse(grammar, 'a b')
    assert ast == ('a', 'b')

    ast = parse(grammar, 'a y')
    assert ast == ('a', 'y')

    with pytest.raises(FailedToken, match=r"expecting 'y'"):
        ast = parse(grammar, 'a c d')
        assert ast == ('a', 'c', 'd')



import sys

from tatsu.mixins.indent import IndentPrintMixin
from tatsu.model import Node
from tatsu.walkers import NodeWalker

THIS_MODULE = sys.modules[__name__]


class PostfixCodeGenerator(NodeWalker, IndentPrintMixin):
    def walk_Add(self, node: Node, *args, **kwargs):
        with self.indent():
            self.walk(node.left)  # type: ignore[attr-defined]
            self.walk(node.right)  # type: ignore[attr-defined]
            self.print('ADD')

    def walk_Subtract(self, node: Node, *args, **kwargs):
        with self.indent():
            self.walk(node.left)  # type: ignore[attr-defined]
            self.walk(node.right)  # type: ignore[attr-defined]
            self.print('SUB')

    def walk_Multiply(self, node: Node, *args, **kwargs):
        with self.indent():
            self.walk(node.left)  # type: ignore[attr-defined]
            self.walk(node.right)  # type: ignore[attr-defined]
            self.print('MUL')

    def walk_Divide(self, node: Node, *args, **kwargs):
        with self.indent():
            self.walk(node.left)  # type: ignore[attr-defined]
            self.walk(node.right)  # type: ignore[attr-defined]
            self.print('DIV')

    def walk_int(self, node: Node, *args, **kwargs):
        self.print('PUSH', node)



import unittest
from ast import parse

from tatsu.exceptions import FailedParse
from tatsu.ngcodegen import pythongen
from tatsu.tool import compile
from tatsu.util import trim


class JoinTests(unittest.TestCase):
    def test_positive_join(self):
        grammar = """
            start = ','%{'x' 'y'}+ ;
        """

        grammar2 = """
            start = (','%{'x'}+|{}) ;
        """

        grammar3 = """
            start = [','%{'x'}+] ;
        """

        model = compile(grammar, 'test')
        pythongen(model)
        ast = model.parse('x y, x y', nameguard=False)
        self.assertEqual([['x', 'y'], ',', ['x', 'y']], ast)
        ast = model.parse('x y x y', nameguard=False)
        self.assertEqual([['x', 'y']], ast)
        try:
            ast = model.parse('y x', nameguard=False)
            self.fail('closure not positive')
        except FailedParse:
            pass

        model = compile(grammar2, 'test')
        ast = model.parse('y x', nameguard=False)
        self.assertEqual([], ast)
        ast = model.parse('x', nameguard=False)
        self.assertEqual(['x'], ast)
        ast = model.parse('x,x', nameguard=False)
        self.assertEqual(['x', ',', 'x'], ast)

        model = compile(grammar3, 'test')
        ast = model.parse('y x', nameguard=False)
        self.assertEqual(None, ast)

    def test_normal_join(self):
        grammar = """
            start = ','%{'x' 'y'} 'z' ;
        """

        model = compile(grammar, 'test')
        pythongen(model)

        ast = model.parse('x y, x y z', nameguard=False)
        self.assertEqual(([['x', 'y'], ',', ['x', 'y']], 'z'), ast)

        ast = model.parse('x y z', nameguard=False)
        self.assertEqual(([['x', 'y']], 'z'), ast)

        ast = model.parse('z', nameguard=False)
        self.assertEqual(([], 'z'), ast)

    def test_group_join(self):
        grammar = """
            start = ('a' 'b')%{'x'}+ ;
        """
        model = compile(grammar, 'test')
        c = pythongen(model)
        parse(c)

        ast = model.parse('x a b x', nameguard=False)
        self.assertEqual(['x', ['a', 'b'], 'x'], ast)

    def test_positive_gather(self):
        grammar = """
            start = ','.{'x' 'y'}+ ;
        """

        grammar2 = """
            start = (','.{'x'}+|{}) ;
        """

        grammar3 = """
            start = [','.{'x'}+] ;
        """

        model = compile(grammar, 'test')
        pythongen(model)
        ast = model.parse('x y, x y', nameguard=False)
        self.assertEqual([['x', 'y'], ['x', 'y']], ast)
        ast = model.parse('x y x y', nameguard=False)
        self.assertEqual([['x', 'y']], ast)
        try:
            ast = model.parse('y x', nameguard=False)
            self.fail('closure not positive')
        except FailedParse:
            pass

        model = compile(grammar2, 'test')
        ast = model.parse('y x', nameguard=False)
        self.assertEqual([], ast)
        ast = model.parse('x', nameguard=False)
        self.assertEqual(['x'], ast)
        ast = model.parse('x,x', nameguard=False)
        self.assertEqual(['x', 'x'], ast)

        model = compile(grammar3, 'test')
        ast = model.parse('y x', nameguard=False)
        self.assertEqual(None, ast)

    def test_normal_gather(self):
        grammar = """
            start = ','.{'x' 'y'} 'z' ;
        """

        model = compile(grammar, 'test')
        pythongen(model)

        ast = model.parse('x y, x y z', nameguard=False)
        self.assertEqual(([['x', 'y'], ['x', 'y']], 'z'), ast)

        ast = model.parse('x y z', nameguard=False)
        self.assertEqual(([['x', 'y']], 'z'), ast)

        ast = model.parse('z', nameguard=False)
        self.assertEqual(([], 'z'), ast)

    def test_group_gather(self):
        grammar = """
            start = ('a' 'b').{'x'}+ ;
        """
        model = compile(grammar, 'test')
        c = pythongen(model)
        parse(c)

        ast = model.parse('x a b x', nameguard=False)
        self.assertEqual(['x', 'x'], ast)

    def test_left_join(self):
        grammar = r"""
            start
                =
                (op)<{number}+ $
                ;


            op
                =
                '+' | '-'
                ;


            number
                =
                /\d+/
                ;
        """
        text = '1 + 2 - 3 + 4'

        model = compile(grammar, 'test')
        self.assertEqual(trim(grammar).strip(), str(model).strip())
        pythongen(model)

        ast = model.parse(text)
        self.assertEqual(('+', ('-', ('+', '1', '2'), '3'), '4'), ast)

    def test_right_join(self):
        grammar = r"""
            start
                =
                (op)>{number}+ $
                ;


            op
                =
                '+' | '-'
                ;


            number
                =
                /\d+/
                ;
        """
        text = '1 + 2 - 3 + 4'

        model = compile(grammar, 'test')
        self.assertEqual(trim(grammar).strip(), str(model).strip())
        pythongen(model)

        ast = model.parse(text)
        self.assertEqual(('+', '1', ('-', '2', ('+', '3', '4'))), ast)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import functools
from collections import defaultdict
from collections.abc import Callable, Collection, Mapping
from copy import copy
from dataclasses import field
from itertools import takewhile
from pathlib import Path
from typing import Any

from .ast import AST
from .contexts import ParseContext
from .exceptions import FailedRef, GrammarError
from .infos import ParserConfig, RuleInfo
from .objectmodel import Node, tatsudataclass
from .util import debug, indent, re, trim, typename
from .util.abctools import chunks, compress_seq

PEP8_LLEN = 72
PRAGMA_RE = r'^\s*#include.*$'

type ffset = set[tuple[str, ...]]


class _ref(str):
    def __repr__(self) -> str:
        return f'<{self}>'


def ref(name: str) -> tuple[str]:
    return (_ref(name),)


def kdot(x: ffset, y: ffset, k: int) -> ffset:
    if not y:
        return {a[:k] for a in x}
    elif not x:
        return {b[:k] for b in y}
    else:
        return {(a + b)[:k] for a in x for b in y}


class ModelContext(ParseContext):
    def __init__(
        self,
        rules,
        /,
        start: str | None = None,
        config: ParserConfig | None = None,
        **settings,
    ):
        config = ParserConfig.new(config, **settings)
        config = config.override(start=start)

        super().__init__(config=config)

        self._rulemap: dict[str, Rule] = {rule.name: rule for rule in rules}

    @property
    def rulemap(self) -> dict[str, Rule]:
        return self._rulemap

    @property
    def pos(self) -> int:
        return self._tokenizer.pos

    def _find_rule(self, name: str) -> Callable:
        return functools.partial(self.rulemap[name]._parse, self)


@tatsudataclass
class Model(Node):
    _lookahead: ffset = field(init=False, default_factory=set)
    _firstset: ffset = field(init=False, default_factory=set)
    _follow_set: ffset = field(init=False, default_factory=set)

    @staticmethod
    def classes() -> list[type]:
        return [
            c
            # Copy globals() before iterating to be thread-safe.
            for c in globals().copy().values()
            if isinstance(c, type) and issubclass(c, Model)
        ]

    def follow_ref(self, rulemap: Mapping[str, Rule]) -> Model:
        return self

    def _parse(self, ctx: ModelContext) -> Any | None:
        ctx.last_node = None
        return None

    def defines(self):
        return []

    def _add_defined_attributes(self, ctx: ModelContext, ast: Any | None = None):
        if ast is None:
            return
        if not hasattr(ast, '_define'):
            return

        defines = dict(compress_seq(self.defines()))

        keys = [k for k, list in defines.items() if not list]
        list_keys = [k for k, list in defines.items() if list]
        ctx._define(keys, list_keys)
        if isinstance(ast, AST):
            ast._define(keys, list_keys)

    def lookahead(self, k: int = 1) -> ffset:
        if not self._lookahead:
            self._lookahead = kdot(self.firstset(k), self.followset(k), k)
        return self._lookahead

    def lookahead_str(self) -> str:
        return ' '.join(sorted(repr(f[0]) for f in self.lookahead() if f))

    def firstset(self, k: int = 1) -> ffset:
        if not self._firstset:
            self._firstset = self._first(k, defaultdict(set))
        return self._firstset

    def followset(self, k: int = 1) -> ffset:
        return self._follow_set

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        return set()

    def _used_rule_names(self):
        return set()

    def _first(self, k: int, f: Mapping[str, ffset]) -> ffset:
        return set()

    def _follow(self, k, fl, a):
        return a

    def is_nullable(self, rulemap: Mapping[str, Rule] | None = None) -> bool:
        return self._nullable()

    def _nullable(self) -> bool:
        return False

    # list of Model that can be invoked at the same position
    def callable_at_same_pos(
        self,
        rulemap: Mapping[str, Rule] | None = None,
    ) -> list[Model]:
        return []

    def nodecount(self) -> int:
        return 1

    def pretty(self, lean: bool = False) -> str:
        return self._pretty(lean=lean)

    def pretty_lean(self):
        return self._pretty(lean=True)

    def _pretty(self, lean=False):
        return f'{typename(self)}: {id(self)}'

    def __str__(self):
        return self.pretty()


@tatsudataclass
class Void(Model):
    def _parse(self, ctx):
        return ctx._void()

    def _pretty(self, lean=False):
        return '()'

    def _nullable(self) -> bool:
        return True


@tatsudataclass
class Dot(Model):
    def _parse(self, ctx):
        return ctx._dot()

    def _pretty(self, lean=False):
        return '/./'

    def _first(self, k: int, f: Mapping[str, ffset]) -> ffset:
        return {('.',)}


@tatsudataclass
class Fail(Model):
    def _parse(self, ctx):
        return ctx._fail()

    def _pretty(self, lean=False):
        return '!()'


@tatsudataclass
class Comment(Model):
    comment: str = ''

    def _pretty(self, lean: bool = False):
        return f'(* {self.comment} *)'


@tatsudataclass
class EOLComment(Comment):
    def _pretty(self, lean=False):
        return f'  # {self.comment}\n'


@tatsudataclass
class EOF(Model):
    def _parse(self, ctx):
        ctx._check_eof()

    def _pretty(self, lean=False):
        return '$'


@tatsudataclass
class Decorator(Model):
    name: str = field(default='')  # note: pre-define as common attribute
    exp: Model = field(default_factory=Void)

    def __post_init__(self):
        super().__post_init__()
        if self.exp is None or not isinstance(self.ast, AST):
            self.exp = self.ast
        self.exp = self.exp or Void()
        assert self.exp is not None, f'{typename(self)}({self.exp})'
        assert self.exp, repr(self)

    def _parse(self, ctx) -> Any | None:
        return self.exp._parse(ctx)

    def defines(self):
        return self.exp.defines()

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        assert isinstance(self.exp, Model), f'{self!r}:{self.exp=!r}'
        return self.exp.missing_rules(rulenames)

    def _used_rule_names(self):
        return self.exp._used_rule_names()

    def _first(self, k: int, f: Mapping[str, ffset]) -> ffset:
        return self.exp._first(k, f)

    def _follow(self, k, fl, a):
        return self.exp._follow(k, fl, a)

    def nodecount(self) -> int:
        return 1 + self.exp.nodecount()

    def _pretty(self, lean=False):
        return self.exp._pretty(lean=lean)

    def _nullable(self) -> bool:
        return self.exp._nullable()

    def callable_at_same_pos(
        self,
        rulemap: Mapping[str, Rule] | None = None,
    ) -> list[Model]:
        return [self.exp]


# NOTE: backwards compatibility
_Decorator = Decorator


@tatsudataclass
class Group(Decorator):
    def _parse(self, ctx):
        with ctx._group():
            self.exp._parse(ctx)
            return ctx.last_node

    def _pretty(self, lean=False):
        exp = self.exp._pretty(lean=lean)
        if len(exp.splitlines()) <= 1:
            return f'({trim(exp)})'
        return f'(\n{indent(exp)}\n)'


@tatsudataclass
class Token(Model):
    token: str = ''

    def __post_init__(self):
        super().__post_init__()
        self.token = self.token or self.ast

    def _parse(self, ctx):
        return ctx._token(self.token)

    def _first(self, k, f) -> ffset:
        return {(self.token,)}

    def _pretty(self, lean=False):
        return repr(self.token)


@tatsudataclass
class Constant(Model):
    literal: str = ''

    def __post_init__(self):
        super().__post_init__()
        self.literal = self.literal or self.ast

    def _parse(self, ctx):
        return ctx._constant(self.literal)

    def _first(self, k, f) -> ffset:
        return {()}

    def _pretty(self, lean=False):
        return f'`{self.literal!r}`'

    def _nullable(self) -> bool:
        return True


@tatsudataclass
class Alert(Constant):
    level: int = 0

    def __post_init__(self):
        super().__post_init__()
        self.literal = self.ast.message.literal
        self.level = len(self.ast.level)

    def _parse(self, ctx):
        return super()._parse(ctx)

    def _pretty(self, lean=False):
        return f'{"^" * self.level}{super()._pretty()}'


@tatsudataclass
class Pattern(Model):
    pattern: str = ''

    def __post_init__(self):
        super().__post_init__()
        self._patterns = self.ast if isinstance(self.ast, list) else [self.ast]
        self.pattern = ''.join(self._patterns)
        self._regex = re.compile(self.pattern)

    def _parse(self, ctx):
        return ctx._pattern(self.pattern)

    def _first(self, k, f) -> ffset:
        x = str(self)
        if bool(self._regex.match('')):
            return {(), (x,)}
        else:
            return {(x,)}

    def _pretty(self, lean=False):
        parts = []
        for pat in (str(p) for p in self._patterns):
            if '/' in pat:
                newpat = pat.replace('"', r'\"')
                regex = f'?"{newpat}"'
            else:
                regex = f'/{pat}/'
            parts.append(regex)
        return '\n+ '.join(parts)

    def _nullable(self) -> bool:
        return bool(self._regex.match(''))

    def __str__(self):
        return self.pattern


@tatsudataclass
class Lookahead(Decorator):
    def _parse(self, ctx):
        with ctx._if():
            return super()._parse(ctx)

    def _pretty(self, lean=False):
        return '&' + self.exp._pretty(lean=lean)

    def _nullable(self) -> bool:
        return True


@tatsudataclass
class NegativeLookahead(Decorator):
    def _parse(self, ctx):
        with ctx._ifnot():
            return super()._parse(ctx)

    def _pretty(self, lean=False):
        return '!' + str(self.exp._pretty(lean=lean))

    def _nullable(self) -> bool:
        return True


@tatsudataclass
class SkipTo(Decorator):
    def _parse(self, ctx):
        super_parse = super()._parse
        return ctx._skip_to(lambda: super_parse(ctx))

    def _first(self, k, f) -> ffset:
        return {('.',)} | super()._first(k, f)

    def _pretty(self, lean=False):
        return '->' + self.exp._pretty(lean=lean)


@tatsudataclass
class Sequence(Model):
    sequence: list[Model] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        if not self.sequence:
            self.sequence = self.ast or []
        assert isinstance(self.sequence, list), self.sequence

    def _parse(self, ctx):
        ctx.last_node = [s._parse(ctx) for s in self.sequence]
        return ctx.last_node

    def defines(self):
        return [d for s in self.sequence for d in s.defines()]

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        return set().union(*[s.missing_rules(rulenames) for s in self.sequence])

    def _used_rule_names(self):
        return set().union(*[s._used_rule_names() for s in self.sequence])

    def _first(self, k, f) -> ffset:
        result = {()}
        for s in self.sequence:
            x = s._first(k, f)
            result = kdot(result, x, k)
        self._firstset = result
        return result

    def _follow(self, k, fl, a):
        fs = a
        for x in reversed(self.sequence):
            if isinstance(x, Call):
                fl[x.name] |= fs
            x._follow(k, fl, fs)
            fs = kdot(x.firstset(k=k), fs, k)
        return a

    def nodecount(self) -> int:
        return 1 + sum(s.nodecount() for s in self.sequence)

    def _pretty(self, lean=False):
        seq = [str(s._pretty(lean=lean)) for s in self.sequence]
        single = ' '.join(seq)
        if len(single) <= PEP8_LLEN and len(single.splitlines()) <= 1:
            return single
        else:
            return '\n'.join(seq)

    def _nullable(self) -> bool:
        return all(s._nullable() for s in self.sequence)

    def callable_at_same_pos(
        self,
        rulemap: Mapping[str, Rule] | None = None,
    ) -> list[Model]:
        head = list(takewhile(lambda c: c.is_nullable(rulemap), self.sequence))
        if len(head) < len(self.sequence):
            head.append(self.sequence[len(head)])
        return head


@tatsudataclass
class Choice(Model):
    options: list[Model] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        self.options = self.options or self.ast
        assert isinstance(self.options, list), repr(self.options)

    def _parse(self, ctx):
        with ctx._choice():
            for o in self.options:
                with ctx._option():
                    ctx.last_node = o._parse(ctx)
                    return ctx.last_node

            lookahead = self.lookahead_str()
            if lookahead:
                raise ctx.newexcept(f'expecting one of: {lookahead}:')
            raise ctx.newexcept('no available options')

    def defines(self):
        return [d for o in self.options for d in o.defines()]

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        return set().union(*[o.missing_rules(rulenames) for o in self.options])

    def _used_rule_names(self):
        return set().union(*[o._used_rule_names() for o in self.options])

    def _first(self, k, f) -> ffset:
        result = set()
        for o in self.options:
            result |= o._first(k, f)
        self._firstset = result
        return result

    def _follow(self, k, fl, a):
        for o in self.options:
            o._follow(k, fl, a)
        return a

    def nodecount(self) -> int:
        return 1 + sum(o.nodecount() for o in self.options)

    def _pretty(self, lean=False):
        options = [str(o._pretty(lean=lean)) for o in self.options]

        multi = any(len(o.splitlines()) > 1 for o in options)
        single = ' | '.join(o for o in options)

        if multi:
            return '\n|\n'.join(indent(o) for o in options)
        elif options and len(single) > PEP8_LLEN:
            return '| ' + '\n| '.join(o for o in options)
        else:
            return single

    def _nullable(self) -> bool:
        return any(o._nullable() for o in self.options)

    def callable_at_same_pos(
        self,
        rulemap: Mapping[str, Rule] | None = None,
    ) -> list[Model]:
        return self.options


@tatsudataclass
class Option(Decorator):
    def _parse(self, ctx):
        result = super()._parse(ctx)
        self._add_defined_attributes(ctx, result)
        return result


@tatsudataclass
class Closure(Decorator):
    def _parse(self, ctx):
        return ctx._closure(lambda: self.exp._parse(ctx))

    def _first(self, k, f) -> ffset:
        efirst = self.exp._first(k, f)
        result = {()}
        for _i in range(k):
            result = kdot(result, efirst, k)
        return {()} | result

    def _pretty(self, lean=False):
        sexp = str(self.exp._pretty(lean=lean))
        if len(sexp.splitlines()) <= 1:
            return f'{{{sexp}}}'
        else:
            return f'{{\n{indent(sexp)}\n}}'

    def _nullable(self) -> bool:
        return True


@tatsudataclass
class PositiveClosure(Closure):
    def _parse(self, ctx):
        return ctx._positive_closure(lambda: self.exp._parse(ctx))

    def _first(self, k, f) -> ffset:
        return self.exp._first(k, f)

    def _pretty(self, lean=False):
        return super()._pretty(lean=lean) + '+'

    def _nullable(self) -> bool:
        return self.exp._nullable()


@tatsudataclass
class Join(Decorator):
    JOINOP = '%'

    sep: Model = Model()

    def __post_init__(self):
        super().__post_init__()
        assert self.sep == self.ast.sep, self.sep

    def _parse(self, ctx):
        def sep():
            return self.sep._parse(ctx)

        def exp():
            return self.exp._parse(ctx)

        return self._do_parse(ctx, exp, sep)

    def _do_parse(self, ctx, exp, sep):
        return ctx._join(exp, sep)

    def _pretty(self, lean=False):
        ssep = self.sep._pretty(lean=lean)
        sexp = str(self.exp._pretty(lean=lean))
        if len(sexp.splitlines()) <= 1:
            return f'{ssep}{self.JOINOP}{{{sexp}}}'
        else:
            return f'{ssep}{self.JOINOP}{{\n{sexp}\n}}'

    def _nullable(self) -> bool:
        return True


class PositiveJoin(Join):
    def _first(self, k, f) -> ffset:
        return self.exp._first(k, f)

    def _do_parse(self, ctx, exp, sep):
        return ctx._positive_join(exp, sep)

    def _pretty(self, lean=False):
        return super()._pretty(lean=lean) + '+'

    def _nullable(self) -> bool:
        return self.exp._nullable()


class LeftJoin(PositiveJoin):
    JOINOP = '<'

    def _do_parse(self, ctx, exp, sep):
        return ctx._left_join(exp, sep)


class RightJoin(PositiveJoin):
    JOINOP = '>'

    def _do_parse(self, ctx, exp, sep):
        return ctx._right_join(exp, sep)


class Gather(Join):
    JOINOP = '.'

    def _do_parse(self, ctx, exp, sep):
        return ctx._gather(exp, sep)


class PositiveGather(Gather):
    def _first(self, k, f) -> ffset:
        return self.exp._first(k, f)

    def _do_parse(self, ctx, exp, sep):
        return ctx._positive_gather(exp, sep)

    def _pretty(self, lean=False):
        return super()._pretty(lean=lean) + '+'

    def _nullable(self) -> bool:
        return self.exp._nullable()


@tatsudataclass
class EmptyClosure(Model):
    def _parse(self, ctx):
        return ctx._empty_closure()

    def _first(self, k, f) -> ffset:
        return {()}

    def _pretty(self, lean=False):
        return '{}'

    def _nullable(self) -> bool:
        return True


@tatsudataclass
class Optional(Decorator):
    def _parse(self, ctx):
        ctx.last_node = None
        self._add_defined_attributes(ctx, ctx.ast)
        with ctx._optional():
            return self.exp._parse(ctx)

    def _first(self, k, f) -> ffset:
        return set({()}) | self.exp._first(k, f)

    def _pretty(self, lean=False):
        exp = self.exp._pretty(lean=lean)
        if len(exp.splitlines()) <= 1:
            return f'[{exp}]'
        return f'[\n{exp}\n]'

    def _nullable(self) -> bool:
        return True


@tatsudataclass
class Cut(Model):
    def _parse(self, ctx):
        ctx._cut()

    def _first(self, k, f) -> ffset:
        return {()}

    def _pretty(self, lean=False):
        return '~'

    def _nullable(self) -> bool:
        return True


@tatsudataclass
class Named(Decorator):
    name: str = field(default='')

    def __post_init__(self):
        if not self.ast:
            self.ast = AST(name=self.name, exp=self.exp)
        super().__post_init__()
        if not self.name:
            raise TypeError(f'{typename(self)}.name is required')
        assert getattr(self, 'name', None) is not None

    def _parse(self, ctx):
        value = self.exp._parse(ctx)
        ctx.ast[self.name] = value
        return value

    def defines(self):
        return [(self.name, False), *super().defines()]

    def _pretty(self, lean=False):
        if lean:
            return self.exp._pretty(lean=True)
        return f'{self.name}:{self.exp._pretty(lean=lean)}'


@tatsudataclass
class NamedList(Named):
    def _parse(self, ctx):
        value = self.exp._parse(ctx)
        ctx.ast._setlist(self.name, value)
        return value

    def defines(self):
        return [(self.name, True), *super().defines()]

    def _pretty(self, lean=False):
        if lean:
            return self.exp._pretty(lean=True)
        return f'{self.name}+:{self.exp._pretty(lean=lean)!s}'


@tatsudataclass
class Override(Named):
    def __post_init__(self):
        # HACK:
        #   The rule for override in the bootstrap grammar uses `@:term`.
        #   That's too difficult to change... So we path conformance with Named!
        #   BaseNode.__post_init__() transfers AST entries to attributes

        self.ast = AST(name='@', exp=self.ast)
        super().__post_init__()

    def defines(self):
        return self.exp.defines()


@tatsudataclass
class OverrideList(NamedList):
    def __post_init__(self):
        self.ast = AST(name='@', exp=self.ast)
        super().__post_init__()

    def defines(self):
        return self.exp.defines()


@tatsudataclass
class Call(Model):
    name: str = ''

    def __post_init__(self):
        if not self.name:
            self.ast = AST(name=self.ast)
        super().__post_init__()
        assert isinstance(self.name, str), self.name

    def follow_ref(self, rulemap: Mapping[str, Rule]) -> Model:
        return rulemap.get(self.name, self)

    def _parse(self, ctx):
        try:
            rule = ctx._find_rule(self.name)
            return rule()
        except KeyError as e:
            raise ctx.newexcept(self.name, exclass=FailedRef) from e

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        if self.name not in rulenames:
            return set({self.name})
        return set()

    def _used_rule_names(self) -> set[str]:
        return {self.name}

    def _first(self, k, f) -> ffset:
        return f[self.name] | {ref(self.name)}

    def _follow(self, k, fl, a):
        fl[self.name] |= a
        return set(a) | {self.name}

    def _pretty(self, lean=False):
        return self.name

    def is_nullable(self, rulemap: Mapping[str, Rule] | None = None) -> bool:
        if rulemap is None:
            return False
        else:
            return rulemap[self.name].is_nullable(rulemap)


@tatsudataclass
class Rule(Decorator):
    params: tuple[str, ...] = field(default_factory=tuple)
    kwparams: dict[str, Any] = field(default_factory=dict)
    decorators: list[str] = field(default_factory=list)
    base: str | None = None
    is_name: bool = False
    is_leftrec: bool = False
    is_memoizable: bool = True

    def __post_init__(self):
        super().__post_init__()
        self.is_name = 'name' in self.decorators
        self.params = self.params or ()

        if not self.kwparams:
            self.kwparams = {}
        elif not isinstance(self.kwparams, dict):
            self.kwparams = dict(self.kwparams)
        self.exp = self.exp if self.exp is not None else Void()
        assert isinstance(self.kwparams, dict), f'{typename(self)}: {self.kwparams=!r}'

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        if self.exp is None:
            debug(f'{self!r} weird with {self.exp=!r}')
        if self.exp.missing_rules is None:
            debug(f'{self!r}\nweird with {self.exp.missing_rules=!r}\n{self.exp=!r}')
        return self.exp.missing_rules(rulenames)

    def _parse(self, ctx):
        return self._parse_rhs(ctx, self.exp)

    def _parse_rhs(self, ctx, exp):
        ruleinfo = RuleInfo(
            self.name,
            exp._parse,
            self.is_leftrec,
            self.is_memoizable,
            self.is_name,
            self.params,
            self.kwparams,
        )
        return ctx._call(ruleinfo)

    def _first(self, k, f) -> ffset:
        self._firstset = self.exp._first(k, f) | f[self.name]
        return self._firstset

    def _follow(self, k, fl, a):
        return self.exp._follow(k, fl, fl[self.name])

    def _nullable(self) -> bool:
        return self.exp._nullable()

    @staticmethod
    def param_repr(p):
        if isinstance(p, int | float) or (isinstance(p, str) and p.isalnum()):
            return str(p)
        else:
            return repr(p)

    def _pretty(self, lean=False):
        str_template = """\
                {is_name}{name}{base}{params}
                    =
                {exp}
                    ;
                """

        if lean:
            params = ''
        else:
            params = (
                ', '.join(self.param_repr(p) for p in self.params)
                if self.params
                else ''
            )

            kwparams = ''
            if self.kwparams:
                kwparams = ', '.join(
                    f'{k}={self.param_repr(v)}' for (k, v) in self.kwparams.items()
                )

            if params and kwparams:
                params = f'({params}, {kwparams})'
            elif kwparams:
                params = f'({kwparams})'
            elif params:
                params = f'::{params}' if len(self.params) == 1 else f'({params})'

        base = f' < {self.base!s}' if self.base else ''

        return trim(str_template).format(
            name=self.name,
            base=base,
            params=params,
            exp=indent(self.exp._pretty(lean=lean)),
            is_name='@name\n' if self.is_name else '',
        )


@tatsudataclass
class RuleInclude(Decorator):
    rule: Rule = field(default_factory=Rule)

    def __post_init__(self):
        super().__post_init__()
        assert isinstance(self.rule, Rule)
        # note: self.ast: str is the rule name
        self.exp = self.rule.exp

    def _pretty(self, lean=False):
        return f'>{self.rule.name}'


@tatsudataclass
class BasedRule(Rule):
    baserule: Rule = field(default_factory=Rule)
    rhs: Model = Model()

    def __post_init__(self):
        super().__post_init__()

        assert isinstance(
            self.baserule,
            Rule,
        ), f'{typename(self.base)}: {self.basrulee=!r}'

        self.params = self.params or self.baserule.params
        self.kwparams = self.kwparams or self.baserule.kwparams
        self.rhs = Sequence(ast=[self.baserule.exp, self.exp])

    def _parse(self, ctx):
        return self._parse_rhs(ctx, self.rhs)

    def defines(self):
        return super().defines() + self.rhs.defines()


@tatsudataclass
class Grammar(Model):
    def __init__(
        self,
        name,
        rules,
        /,
        config: ParserConfig | None = None,
        directives: dict | None = None,
        **settings,
    ):
        super().__init__()
        assert isinstance(rules, list), str(rules)
        directives = directives or {}
        self.directives = directives

        config = ParserConfig.new(config=config, **settings)
        config = config.hard_override(**directives)
        self.config = config

        self.rules: list[Rule] = rules
        self._rulemap: dict[str, Rule] = {rule.name: rule for rule in rules}

        if name is None:
            name = self.directives.get('grammar')
        if name is None:
            name = self.config.name
        if name is None and config.filename is not None:
            name = Path(config.filename).stem
        if name is None:
            name = 'My'
        self.name = name

        missing: set[str] = self.missing_rules(set(self._rulemap))
        if missing:
            msg = '\n'.join(['', *sorted(missing)])
            raise GrammarError('Unknown rules, no parser generated:' + msg)

        self._calc_lookahead_sets()

    @property
    def rulemap(self) -> dict[str, Rule]:
        return self._rulemap

    @property
    def keywords(self) -> Collection[str]:
        return self.config.keywords

    @property
    def semantics(self) -> type[object] | None:
        return self.config.semantics

    @semantics.setter
    def semantics(self, value: type[object]):
        self.config.semantics = value

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        return set().union(*[rule.missing_rules(rulenames) for rule in self.rules])

    def _used_rule_names(self) -> set[str]:
        if not self.rules:
            return set()

        used = {'start', self.rules[0].name}
        prev: set[str] = set()
        while used != prev:
            prev = used
            used |= set().union(
                *[rule._used_rule_names() for rule in self.rules if rule.name in used],
            )
        return used

    def used_rules(self) -> list[Rule]:
        used = self._used_rule_names()
        return [rule for rule in self.rules if rule.name in used]

    @property
    def first_sets(self):
        return self._firstset

    def _calc_lookahead_sets(self, k: int = 1):
        self._calc_first_sets(k=k)
        self._calc_follow_sets(k=k)

    def _calc_first_sets(self, k: int = 1):
        f: dict[str, ffset] = defaultdict(set)
        f1 = None
        while f1 != f:
            f1 = copy(f)
            for rule in self.rules:
                f[rule.name] |= rule._first(k, f)

        # cache results
        for rule in self.rules:
            rule._firstset = f[rule.name]

    def _calc_follow_sets(self, k: int = 1):
        fl: dict[str, ffset] = defaultdict(set)
        fl1 = None
        while fl1 != fl:
            fl1 = copy(fl)
            for rule in self.rules:
                rule._follow(k, fl, set())

        # cache results
        for rule in self.rules:
            rule._follow_set = fl[rule.name]

    def parse(
        self,
        text: str,
        /,
        *,
        ctx: ParseContext | None = None,
        config: ParserConfig | None = None,
        **settings,
    ):
        config = self.config.override_config(config)
        # note: bw-comp: allow overriding directives
        config = config.override(**settings)

        if isinstance(config.semantics, type):
            raise TypeError(
                'semantics must be an object instance or None, '
                f'not class {config.semantics!r}',
            )

        start = config.effective_start_rule_name()
        if start is None:
            start = self.rules[0].name
            config.start = start
            config.start_rule = None
            config.rule_name = None

        if ctx is None:
            ctx = ModelContext(self.rules, config=config)
        return ctx.parse(text, config=config)

    def nodecount(self) -> int:
        return 1 + sum(r.nodecount() for r in self.rules)

    def _pretty(self, lean: bool = False) -> str:
        regex_directives = {'comments', 'eol_comments', 'whitespace'}
        str_directives = {'comments', 'grammar'}
        string_directives = {'namechars'}

        directives = ''
        for directive, value in self.directives.items():
            fmt = dict(  # noqa: C408
                name=directive,
                frame='/' if directive in regex_directives else '',
                value=(
                    repr(value)
                    if directive in string_directives
                    else str(value) if directive in str_directives else value
                ),
            )
            directives += '@@{name} :: {frame}{value}{frame}\n'.format(**fmt)
        if directives:
            directives += '\n'

        keywords = '\n'.join(
            '@@keyword :: ' + ' '.join(repr(k) for k in c if k is not None)
            for c in chunks(sorted(self.keywords), 8)
        ).strip()
        keywords = '\n\n' + keywords + '\n' if keywords else ''

        rules = (
            '\n\n'.join(str(rule._pretty(lean=lean)) for rule in self.rules)
        ).rstrip() + '\n'
        return directives + keywords + rules



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
The Renderer class provides the infrastructure for generating template-based
code. It's used by the .grammars module for parser generation.
"""

from __future__ import annotations

import itertools
import string

from ..util import indent, trim
from ..util.abctools import isiter


def render(item, join='', **fields):
    """Render the given item"""
    if item is None:
        return ''
    elif isinstance(item, str):
        return item
    elif isinstance(item, Renderer):
        return item.render(join=join, **fields)
    elif isiter(item):
        return join.join(render(e, **fields) for e in iter(item) if e is not None)
    elif isinstance(item, int | float):
        return item
    else:
        return str(item)


class RenderingFormatter(string.Formatter):
    def render(self, item, join='', **fields):
        return render(item, join=join, **fields)

    def format_field(self, value, format_spec):
        if ':' not in format_spec:
            return super().format_field(self.render(value), format_spec)

        ind, sep, fmt = format_spec.split(':')
        if sep == '\\n':
            sep = '\n'

        if not ind:
            ind = 0
            mult = 0
        elif '*' in ind:
            ind, mult = ind.split('*')
        else:
            mult = 4
        ind = int(ind)
        mult = int(mult)

        if not fmt:
            fmt = '%s'

        if isiter(value):
            return indent(sep.join(fmt % self.render(v) for v in value), ind, mult)
        else:
            return indent(fmt % self.render(value), ind, mult)


class Renderer:
    """Renders the fileds in the current object using a template
    provided statically, on the constructor, or as a parameter
    to render().

    Fields with a leading underscore are not made available to
    the template. Additional fields may be made available by
    overriding render_fields().
    """

    template = '{__class__}'
    _counter = itertools.count()
    _formatter = RenderingFormatter()

    def __init__(self, template=None):
        if template is not None:
            self.template = template

    @classmethod
    def counter(cls):
        return next(cls._counter)

    @classmethod
    def reset_counter(cls):
        Renderer._counter = itertools.count()

    @property
    def formatter(self):
        return self._formatter

    @formatter.setter
    def formatter(self, value):
        self._formatter = value

    def rend(self, item, join='', **fields):
        """A shortcut for self._formatter.render()"""
        return self._formatter.render(item, join=join, **fields)

    def indent(self, item, ind=1, multiplier=4):
        return indent(self.rend(item), indent=ind, multiplier=multiplier)

    def trim(self, item, tabwidth=4):
        return trim(self.rend(item), tabwidth=tabwidth)

    def render_fields(self, fields):
        """Pre-render fields before rendering the template."""
        return

    def render(self, **fields):
        template = fields.pop('template', None)
        fields.update(__class__=self.__class__.__name__)
        fields.update({k: v for k, v in vars(self).items() if not k.startswith('_')})

        override = self.render_fields(fields)  # pylint: disable=assignment-from-none
        if override is not None:
            template = override
        elif template is None:
            template = self.template

        try:
            return self._formatter.format(trim(template), **fields)
        except KeyError as e:
            # find the missing key
            keys = (p[1] for p in self._formatter._parse(template))  # type: ignore
            for key in keys:
                if key and key not in fields:
                    raise KeyError(key, type(self)) from e
            raise

    def __str__(self):
        return self.render()

    def __repr__(self):
        return str(self)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import importlib
import itertools
from pathlib import Path

from .util import misc
from .walkers import NodeWalker

__all__ = ['draw']

# SHAPE           DESCRIPTION
# box             Standard rectangle
# circle          Basic circle
# ellipse         Default shape; an oval
# point           Microscopic filled circle (junctions)
# egg             Asymmetric oval
# triangle        Upward pointing triangle
# plaintext       Text with no border
# diamond         Rhombus (decision points)
# trapezium       Trapezoid
# parallelogram   Slanted rectangle (I/O or rules)
# house           Polygon shaped like a house
# pentagon        Five-sided polygon
# hexagon         Six-sided polygon
# septagon        Seven-sided polygon
# octagon         Eight-sided polygon
# doublecircle    Circle with an inner ring
# doubleoctagon   Octagon with an inner border
# tripleoctagon   Octagon with two inner borders
# invtriangle     Upside-down triangle
# invtrapezium    Upside-down trapezoid
# invhouse        Upside-down house shape
# rect            Alias for box
# rectangle       Alias for box
# square          Box with equal sides
# star            Multi-pointed star
# cylinder        3D-style tube (databases)
# note            Paper with a folded corner
# tab             Folder-style tab
# folder          Classic file folder shape
# box3d           Rectangle with 3D depth
# component       Box with "plug protrusions"
# larrow          Arrow pointing left
# rarrow          Arrow pointing right
# record          Box for structured data fields
# Mrecord         Record with rounded corners
# none            No shape and no label space


def available() -> bool:
    return misc.module_available('graphviz') and misc.platform_has_command('dot')


def draw(filename, grammar):
    traverser = DiagramNodeWalker()
    traverser.walk(grammar)
    traverser.render(filename)


class DiagramNodeWalker(NodeWalker):
    def __init__(self):
        super().__init__()

        # splines=
        # true     ?
        # line     Strictly straight lines between nodes.
        # polyline Straight segments that can change direction at internal points.
        # ortho    All lines are strictly horizontal or vertical (manhattan routing).
        # curved   Smooth Bezier curves (default behavior for splines=true).
        self.graphviz = importlib.import_module('graphviz')

        self.top_graph = self.graphviz.Digraph(
            engine='dot',
            graph_attr={'rankdir': 'LR', 'packMode': 'clust', 'splines': 'true'},
        )
        self.stack = [self.top_graph]
        self.node_count = 0

    @property
    def graph(self):
        return self.stack[-1]

    def render(self, filename):
        filepath = Path(filename)
        fmt = filepath.suffix.lstrip('.') or 'dot'
        fmt = 'jpeg' if fmt == 'jpg' else fmt
        self.graph.render(
            # filename=filename,
            outfile=filename,
            format=fmt,
            cleanup=True,
        )

    def push_graph(self, name=None, **attr):
        if name is None:
            self.node_count += 1
            name = f'cluster_g{self.node_count}'
        elif not name.startswith('cluster'):
            # graphviz needs "cluster" prefix to draw a bounding box
            name = f'cluster_{name}'

        # In graphviz, subgraphs are created as new objects
        attr['style'] = 'invis'
        new_subgraph = self.graphviz.Digraph(name=name, graph_attr=attr)
        self.stack.append(new_subgraph)
        return new_subgraph

    def pop_graph(self):
        last_graph = self.stack.pop()
        # In graphviz, you must explicitly add the child back to the parent
        self.graph.subgraph(last_graph)

    def node(self, label, id=None, **attr):
        if id is None:
            self.node_count += 1
            id = f'n{self.node_count}'

        # graphviz .node() takes the ID and then attributes
        self.graph.node(id, label=label, **attr)
        return id  # We return the ID string to be used for edges

    def tnode(self, name, **attr):
        return self.node(name, shape='hexagon', **attr)

    def dot(self):
        return self.node('', shape='point', width='0.01')

    def start_node(self):
        return self.dot()

    def ref_node(self, name):
        return self.node(name, shape='box', style='rounded')

    def rule_node(self, name, **attr):
        # Using the name as the ID for rules as per original logic
        return self.node(name, id=name, shape='box', style='bold', **attr)

    def end_node(self):
        return self.node('', shape='point', width='0.1')

    def edge(self, s, e, **attr):
        # style       Changes the line pattern solid, dashed, dotted, bold, tapered
        # color       Changes the line color red, blue, #FFA500, transparent
        # arrowhead   Changes the shape of the tip normal, dot, none, vee, diamond
        # penwidth    Changes line thickness Numeric value (e.g., 2.0, 0.5)
        # label       Adds text to the edge Any string
        if not (s and e):
            return
        self.graph.edge(s, e, **attr)

    def redge(self, s, e):
        return self.edge(s, e, dir='back')

    def zedge(self, s, e):
        # 'len' is for neato, but kept for parity
        return self.edge(s, e, len='0.000001')

    def nedge(self, s, e):
        return self.edge(s, e, style='invisible', dir='none')

    # --- Walker Logic (Mostly identical, utilizing new wrapper methods) ---

    def walk_decorator(self, d):
        return self.walk(d.exp)

    def walk_default(self, node):
        pass

    def walk_grammar(self, g):
        # Creating clusters for organization
        self.push_graph(name=f"{g.name}_0")
        try:
            vrules = [self.walk(r) for r in reversed(g.rules)]
        finally:
            self.pop_graph()

        if not vrules:
            return None, None

        s, t = vrules[0][0], vrules[-1][1]
        return s, t

    def walk_rule(self, r):
        self.push_graph(name=r.name)
        try:
            i, e = self.walk(r.exp)
            s = self.rule_node(r.name)
            self.edge(s, i)
            t = self.end_node()
            self.edge(e, t)
            return s, t
        finally:
            self.pop_graph()

    def walk_optional(self, o):
        i, e = self.walk_decorator(o)
        ni = self.dot()
        ne = self.dot()
        self.zedge(ni, i)
        self.edge(ni, ne)
        self.zedge(e, ne)
        return ni, ne

    def walk_closure(self, r):
        self.push_graph(rankdir='TB')
        try:
            i, e = self.walk_decorator(r)
            ni = self.dot()
            self.zedge(ni, i)
            self.zedge(e, ni)
            return ni, ni
        finally:
            self.pop_graph()

    def walk_choice(self, c):
        vopt = [self.walk(o) for o in c.options]
        vopt = [o for o in vopt if o is not None]
        ni = self.dot()
        ne = self.dot()
        for i, e in vopt:
            self.edge(ni, i)
            self.edge(e, ne)
        return ni, ne

    def walk_sequence(self, s):
        vseq = [self.walk(x) for x in s.sequence if x is not None]
        vseq = [x for x in vseq if x is not None]

        if not vseq:
            return None, None

        first, _ = vseq[0]
        _, last = vseq[-1]

        for (_, e), (i, _) in itertools.pairwise(vseq):
            self.edge(e, i)

        return first, last

    # Simplified remaining walkers for brevity
    def walk_call(self, rr):
        n = self.ref_node(rr.name)
        return n, n

    def walk_pattern(self, p):
        n = self.tnode(p.pattern)
        return n, n

    def walk_token(self, t):
        n = self.tnode(t.token)
        return n, n

    def walk_eof(self, v):
        n = self.node('$EOF')
        return n, n

    def walk_lookahead(self, v):
        i, e = self.walk(v.exp)
        s = self.node('&', shape='diamond')
        self.zedge(s, i)
        return s, e

    def walk_negative_lookahead(self, v):
        try:
            i, e = self.walk(v.exp)
        except TypeError as e:
            raise TypeError(f'{e} {v.exp}') from e
        s = self.node('&', shape='diamond')
        s = self.node('!', shape='diamond')
        self.zedge(s, i)
        return s, e

    def walk_void(self, v):
        return None, None

    def walk_fail(self, v):
        return None, None

    def ENDRULE(self, ast):
        return None, None

    def EMPTYLINE(self, ast, *args):
        return None, None



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
import warnings
from collections.abc import Collection
from dataclasses import dataclass, field
from typing import Any, override

from .tokenizing import NullTokenizer, Tokenizer
from .util.configs import Config
from .util.misc import cached_re_compile
from .util.undefined import Undefined
from .util.unicode_characters import C_DERIVE

MEMO_CACHE_SIZE = 4 * 1024


@dataclass
class ParserConfig(Config):
    name: str | None = 'Test'
    filename: str = ''

    start: str | None = None

    tokenizercls: type[Tokenizer] = NullTokenizer
    semantics: Any = None

    comment_recovery: bool = False

    memoization: bool = True
    memoize_lookaheads: bool = True
    memo_cache_size: int = MEMO_CACHE_SIZE
    prune_memos_on_cut: bool = True

    colorize: bool = True  # INFO: requires the colorama library
    trace: bool = False
    trace_filename: str = ''
    trace_length: int = 72
    trace_separator: str = C_DERIVE

    # parser directives
    grammar: str | None = None
    left_recursion: bool = True

    comments: str | None = None
    eol_comments: str | None = None
    keywords: Collection[str] = field(default_factory=set)

    ignorecase: bool = False
    namechars: str | None = None
    nameguard: bool | None = None  # implied by namechars
    whitespace: str | None = Undefined  # type: ignore
    parseinfo: bool = False

    # WARNING: DEPRECATED: some old projects use these
    owner: Any = None
    extra: Any = None
    start_rule: str | None = None
    rule_name: str | None = None
    comments_re: re.Pattern | str | None = None
    eol_comments_re: re.Pattern | str | None = None

    def __post_init__(self):  # pylint: disable=W0235
        if self.ignorecase:
            self.keywords = {k.upper() for k in self.keywords}
        super().__post_init__()

        if not self.memoization:
            self.left_recursion = False

        if self.namechars:
            self.nameguard = True

        if isinstance(self.semantics, type):
            raise TypeError(
                f'semantics must be an object instance or None, not class {self.semantics!r}',
            )

        self._deprecate_and_compile_comments()

    def _deprecate_and_compile_comments(self):
        # note: handle deprecations gracefully
        if self.comments_re:
            warnings.warn(
                'ParserConfig.comments_re is deprecated: use `comments`',
                stacklevel=3,
            )
            if not self.comments:
                self.comments = str(self.comments_re)
            del self.comments_re

        if self.eol_comments_re:
            warnings.warn(
                'ParserConfig.eol_comments_re is deprecated: use `eol_comments`',
                stacklevel=3,
            )
            if not self.eol_comments:
                self.eol_comments = str(self.eol_comments_re)
            del self.eol_comments_re

        if self.comments and isinstance(self.comments, str):
            cached_re_compile(self.comments)
        if self.eol_comments and not isinstance(self.eol_comments, re.Pattern):
            cached_re_compile(self.eol_comments)
        if self.whitespace and not isinstance(self.whitespace, re.Pattern):
            cached_re_compile(self.whitespace)

        for name in ('start_rule', 'rule_name'):
            if (value := getattr(self, name, None)) is None:
                continue
            warnings.warn(
                message=(
                    f'\nSetting {name}={value!r} is deprecated.'
                    f' Use start={value!r} for the name of the start rule.'
                ),
                category=DeprecationWarning,
                stacklevel=2,
            )
            break
        self.start = self.effective_start_rule_name()

    def effective_start_rule_name(self):
        # NOTE: there are legacy reasons for this mess
        return (
            self.start
            or getattr(self, 'start_rule', None)
            or getattr(self, 'rule_name', None)
        )

    @override
    def override(self, **settings: Any) -> ParserConfig:
        result = super().override(**settings)
        if 'grammar' in settings:
            result.name = result.grammar
        self._deprecate_and_compile_comments()
        return result



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import difflib
import importlib
import json
import pickle
import pprint
import py_compile
import shutil
import sys
from pathlib import Path

import pytest

from tatsu import diagrams
from tatsu.ngcodegen import pythongen
from tatsu.parser import TatSuParser, TatSuParserGenerator
from tatsu.parser_semantics import TatSuGrammarSemantics
from tatsu.semantics import ASTSemantics
from tatsu.tool import to_python_sourcecode
from tatsu.util import asjson
from tatsu.walkers import DepthFirstWalker

tmp = Path('./tmp').resolve()
sys.path.insert(0, str(tmp))


@pytest.mark.dependency()
def test_00_with_boostrap_grammar():
    print('-' * 20, 'phase 00 - parse using the bootstrap grammar')

    tmp = Path('./tmp')
    if (tmp / '00.ast').is_file():
        shutil.rmtree('./tmp')
    tmp.mkdir(exist_ok=True)
    text = Path('grammar/tatsu.tatsu').read_text()
    g = TatSuParser('TatSuBootstrap')
    grammar0 = g.parse(text, semantics=ASTSemantics(), parseinfo=False)
    ast0 = json.dumps(asjson(grammar0), indent=2)
    Path('./tmp/00.ast').write_text(ast0)


@pytest.mark.dependency('test_00_with_boostrap_grammar')
def test_01_with_parser_generator():
    print('-' * 20, 'phase 01 - parse with parser generator')
    text = Path('grammar/tatsu.tatsu').read_text()
    g = TatSuParserGenerator('TatSuBootstrap')
    result = g.parse(text)
    generated_grammar1 = str(result)
    Path('./tmp/01.tatsu').write_text(generated_grammar1)


@pytest.mark.dependency('test_01_with_parser_generator')
def test_02_previous_output_generator():
    print('-' * 20, 'phase 02 - parse previous output with the parser generator')
    text = Path('grammar/tatsu.tatsu').read_text()
    g = TatSuParserGenerator('TatSuBootstrap')
    result = g.parse(text)
    generated_grammar1 = str(result)

    text = Path('./tmp/01.tatsu').read_text()
    g = TatSuParserGenerator('TatSuBootstrap')
    result = g.parse(text)
    generated_grammar2 = str(result)
    Path('./tmp/02.tatsu').write_text(generated_grammar2)
    assert generated_grammar2 == generated_grammar1


@pytest.mark.dependency('test_02_previous_output_generator')
def test_03_repeat_02():
    print('-' * 20, 'phase 03 - repeat')
    text = Path('./tmp/02.tatsu').read_text()
    g = TatSuParser('TatSuBootstrap')
    ast3 = g.parse(text)
    Path('./tmp/03.ast').write_text(json.dumps(asjson(ast3), indent=2))


@pytest.mark.dependency('test_03_repeat_02')
def test_04_repeat_03():
    print('-' * 20, 'phase 04 - repeat')
    text = Path('./tmp/01.tatsu').read_text()
    g = TatSuParserGenerator('TatSuBootstrap')
    result = g.parse(text)
    generated_grammar2 = str(result)

    text = Path('./tmp/02.tatsu').read_text()
    g = TatSuParserGenerator('TatSuBootstrap')
    parser = g.parse(text)
    generated_grammar4 = str(parser)
    Path('./tmp/04.tatsu').write_text(generated_grammar4)
    assert generated_grammar4 == generated_grammar2


@pytest.mark.dependency('test_04_repeat_03')
def test_05_parse_with_model():
    print('-' * 20, 'phase 05 - parse using the grammar model')
    text = Path('./tmp/02.tatsu').read_text()
    g = TatSuParserGenerator('TatSuBootstrap')
    parser = g.parse(text)

    text = Path('./tmp/04.tatsu').read_text()
    ast5 = parser.parse(text)
    Path('./tmp/05.ast').write_text(json.dumps(asjson(ast5), indent=2))


@pytest.mark.dependency('test_05_parse_with_model')
def test_06_generate_code():
    print('-' * 20, 'phase 06 - generate parser code')
    text = Path('./tmp/02.tatsu').read_text()
    g = TatSuParserGenerator('TatSuBootstrap')
    parser = g.parse(text)
    gencode6 = pythongen(parser)
    Path('./tmp/g06.py').write_text(gencode6)


@pytest.mark.dependency('test_06_generate_code')
def test_07_import_generated_code():
    print('-' * 20, 'phase 07 - import generated code')

    text = Path('grammar/tatsu.tatsu').read_text()
    gencode7 = to_python_sourcecode(text)
    Path('./tmp/g07.py').write_text(gencode7)
    assert Path('./tmp/g07.py').is_file()

    compiled_path = py_compile.compile('./tmp/g07.py', doraise=True)
    if compiled_path is None:
        pytest.fail('Could not compile module')
    assert compiled_path is not None
    compiled_path = Path(compiled_path)
    print('COMPILED', compiled_path)

    tmpdir = Path('./tmp')
    tmpdir.mkdir(parents=True, exist_ok=True)
    assert tmpdir.is_dir()
    Path('./tmp/__init__.py').touch()
    Path('./tmp/__init__.pyc').touch()
    print('CURRENT', Path.cwd())
    importlib.invalidate_caches()
    g07 = importlib.import_module('g07', package='tmp')

    # g07 = __import__('g07')
    assert g07
    generated_parser = g07.TatSuParser
    assert generated_parser


@pytest.mark.dependency('test_07_import_generated_code')
def test_08_compile_with_generated():
    print('-' * 20, 'phase 08 - compile using generated code')
    Path('./tmp').mkdir(parents=True, exist_ok=True)
    Path('./tmp/__init__.py').touch()
    importlib.invalidate_caches()
    g07 = importlib.import_module('g07', package='tmp')
    generated_parser = g07.TatSuParser
    assert generated_parser

    ast0 = Path('./tmp/00.ast').read_text()

    parser = generated_parser(trace=False)
    text = Path('grammar/tatsu.tatsu').read_text()
    result = parser.parse(text, semantics=ASTSemantics(), parseinfo=False)
    ast8 = json.dumps(asjson(result), indent=2)
    Path('./tmp/08.ast').write_text(ast8)
    print('DIFF')
    pprint.pprint(list(difflib.unified_diff(ast0.splitlines(), ast8.splitlines())))
    assert ast0 == ast8


@pytest.mark.dependency('test_08_compile_with_generated')
def test_09_parser_with_semantics():
    print('-' * 20, 'phase 09 - Generate parser with semantics')
    text = Path('grammar/tatsu.tatsu').read_text()
    parser = TatSuParserGenerator('TatSuBootstrap')
    g9 = parser.parse(text)
    generated_grammar9 = str(g9)
    Path('./tmp/09.tatsu').write_text(generated_grammar9)
    text = Path('grammar/tatsu.tatsu').read_text()
    g = TatSuParserGenerator('TatSuBootstrap')
    result = g.parse(text)
    generated_grammar1 = str(result)
    assert generated_grammar9 == generated_grammar1


@pytest.mark.dependency('test_09_parser_with_semantics')
def test_10_with_model_and_semantics():
    print('-' * 20, 'phase 10 - Parse with a model using a semantics')
    text = Path('grammar/tatsu.tatsu').read_text()
    parser = TatSuParserGenerator('TatSuBootstrap')
    g9 = parser.parse(text)
    g10 = g9.parse(
        text,
        start='start',
        semantics=TatSuGrammarSemantics('TatSuBootstrap'),
    )
    generated_grammar10 = str(g10)
    Path('./tmp/10.tatsu').write_text(generated_grammar10)
    gencode10 = pythongen(g10)
    Path('./tmp/g10.py').write_text(gencode10)


@pytest.mark.dependency('test_10_with_model_and_semantics')
def test_11_with_pickle_and_retry():
    print('-' * 20, 'phase 11 - Pickle the model and try again.')
    text = Path('grammar/tatsu.tatsu').read_text()
    parser = TatSuParserGenerator('TatSuBootstrap')
    g9 = parser.parse(text)
    g10 = g9.parse(
        text,
        start='start',
        semantics=TatSuGrammarSemantics('TatSuBootstrap'),
    )
    with Path('./tmp/11.tatsuc').open('wb') as f:
        pickle.dump(g10, f)
    with Path('./tmp/11.tatsuc').open('rb') as f:
        g11 = pickle.load(f)
    r11 = g11.parse(
        text,
        start='start',
        semantics=TatSuGrammarSemantics('TatSuBootstrap'),
    )
    Path('./tmp/11.tatsu').write_text(str(g11))
    gencode11 = pythongen(r11)
    Path('./tmp/bootstrap_g11.py').write_text(gencode11)


@pytest.mark.dependency('test_11_with_pickle_and_retry')
def test_12_walker():
    print('-' * 20, 'phase 12 - Walker')

    class PrintNameWalker(DepthFirstWalker):
        def __init__(self):
            super().__init__()
            self.walked = []

        def walk_default(self, o):
            self.walked.append(o.__class__.__name__)

    with Path('./tmp/11.tatsuc').open('rb') as f:
        g11 = pickle.load(f)
    v = PrintNameWalker()
    v.walk(g11)
    Path('./tmp/12.txt').write_text('\n'.join(v.walked))


@pytest.mark.dependency('test_12_walker')
@pytest.mark.skipif(not diagrams.available(), reason='graphviz is not available')
def test_13_diagram():
    print('-' * 20, 'phase 13 - Graphics')
    with Path('./tmp/11.tatsuc').open('rb') as f:
        g11 = pickle.load(f)
    diagrams.draw('./tmp/13.png', g11)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import enum
import inspect
import json
import weakref
from collections.abc import Mapping
from typing import Any, Protocol, runtime_checkable

from .abctools import as_namedtuple, isiter, rowselect

__all__ = ['AsJSONMixin', 'JSONSerializable', 'asjson', 'asjsons', 'plainjson']


@runtime_checkable
class JSONSerializable(Protocol):
    def __json__(self, seen: set[int] | None = None) -> Any: ...


class AsJSONMixin:
    __slots__ = ()

    def __json__(self, seen: set[int] | None = None) -> Any:
        pub = self.__pub__()
        return {'__class__': type(self).__name__, **asjson(pub, seen=seen)}

    def __pub__(self) -> dict[str, Any]:
        # Gemini (2026-01-26)
        def is_public(name: str, value: Any) -> bool:
            return not (name.startswith('_') or inspect.isroutine(value))

        return rowselect(dir(self), vars(self), where=is_public)


def asjson(obj: Any, seen: set[int] | None = None) -> Any:
    """
    Produce a JSON-serializable version of the input structure.
    """
    # Gemini (2026-01-26)
    seen = seen if seen is not None else set()

    def dfs(node: Any) -> Any:
        if node is None or isinstance(node, int | float | str | bool):
            return node

        node_id = id(node)
        if node_id in seen:
            return f"{type(node).__name__}@0x{hex(node_id).upper()[2:]}"

        seen.add(node_id)
        try:
            match node:
                case JSONSerializable() as serializable:
                    if not isinstance(serializable, type):
                        result = serializable.__json__(seen=seen)
                    else:
                        result = serializable
                case enum.Enum() as en:
                    result = dfs(en.value)
                case _ if isinstance(
                    node,
                    (weakref.ReferenceType, *weakref.ProxyTypes),
                ):
                    result = f"{type(node).__name__}@0x{hex(node_id).upper()[2:]}"
                case _ if nt := as_namedtuple(node):
                    result = dfs(nt._asdict())
                case Mapping() as mapping:
                    result = {str(k): dfs(v) for k, v in mapping.items()}
                case list() | tuple() | set() as seq:
                    result = [dfs(e) for e in seq]
                case _ if isiter(node):
                    result = [dfs(e) for e in node]
                case _:
                    result = repr(node)
            return result
        finally:
            seen.discard(node_id)

    return dfs(obj)


def plainjson(obj: Any) -> Any:
    if isinstance(obj, Mapping):
        return {
            name: plainjson(value)
            for name, value in obj.items()
            if name not in {'__class__', 'parseinfo'}
        }
    elif isinstance(obj, weakref.ReferenceType | weakref.ProxyType):
        return '@ref'
    elif isinstance(obj, str) and obj.startswith('@'):
        return '@ref'
    elif isiter(obj):
        return [plainjson(e) for e in obj]
    else:
        return obj


class FallbackJSONEncoder(json.JSONEncoder):
    """A JSON Encoder that falls back to repr() for non-JSON-aware objects."""

    def default(self, o: Any) -> str:
        return repr(o)


def asjsons(obj: Any) -> str:
    return json.dumps(asjson(obj), indent=2, cls=FallbackJSONEncoder)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import types

import pytest

import tatsu
from tatsu.exceptions import FailedParse

INPUT = """
    1d3
"""
OUTPUT = {'number_of_dice': '1', 'sides': '3'}

GRAMMAR = r"""
    start = expression $;

    int = /-?\d+/ ;

    dice = number_of_dice:factor /d|D/ sides:factor;

    expression = addition ;

    addition
        =
        | left:dice_expr op:('+' | '-') ~ right:addition
        | dice_expr
        ;

    dice_expr
        =
        | dice
        | factor
        ;

    factor
        =
        | '(' ~ @:expression ')'
        | int
        ;
"""


def generate_and_load_parser(name, grammar):
    code = tatsu.to_python_sourcecode(grammar, name='Test')
    print(code)
    module = types.ModuleType(name)
    module.__file__ = '<generated>'
    exec(compile(code, module.__file__, 'exec'), module.__dict__)
    return module.TestParser()


def test_model_parse():
    model = tatsu.compile(name='Test', grammar=GRAMMAR)
    assert model.parse(INPUT) == OUTPUT


def test_codegen_parse():
    parser = generate_and_load_parser('test_codegen_parse', GRAMMAR)
    output = parser.parse(INPUT, parseinfo=False)
    assert output == OUTPUT


def test_error_messages():
    grammar = """
        @@grammar :: ORDER
        alphabet = a b others $ ;

        a = 'a' ;
        b = 'b' ;
        others = 'c' | 'd' | 'e' | 'f' |'g' | 'h' | 'i' | 'j' | 'k' | 'l' | 'm' | 'n' | 'o';
    """
    input = 'a b'

    e1 = None
    model = tatsu.compile(grammar)
    try:
        model.parse(input)
    except FailedParse as e:
        e1 = str(e)
        assert (
            "expecting one of: 'c' 'd' 'e' 'f' 'g' 'h' 'i' 'j' 'k' 'l' 'm' 'n' 'o'"
            in e1
        )


def test_name_checked():
    grammar = r"""
        @@grammar :: Test
        @@ignorecase :: True
        @@keyword :: if

        start = rule ;
        rule = @:word if_exp $ ;
        if_exp = 'if' digit ;
        @name word = /\w+/ ;
        digit = /\d/ ;
    """

    def subtest(parser):
        parser.parse('nonIF if 1', trace=False)
        with pytest.raises(FailedParse):
            parser.parse('if if 1', trace=False)
        with pytest.raises(FailedParse):
            parser.parse('IF if 1', trace=False)

    parser = tatsu.compile(grammar, 'Test')
    subtest(parser)
    parser = generate_and_load_parser('test_name_checked', grammar)
    subtest(parser)


def test_first_rule():
    grammar = """
        @@grammar :: Test

        true = 'test' @: `True` $ ;
        start = 'test' @: `False` $ ;
    """
    parser = tatsu.compile(grammar, 'Test')
    ast = parser.parse('test')
    assert ast is True
    parser = generate_and_load_parser('test_first_rule', grammar)
    ast = parser.parse('test')
    assert ast is True


def test_dynamic_compiled_ast():
    grammar = r"""
        test::Test = 'TEST' ['A' a:number] ['B' b:number] ;
        number::int = /\d+/ ;
    """

    parser = tatsu.compile(grammar)

    code = tatsu.to_python_sourcecode(grammar, name='Test', filename='test.py')
    module = types.ModuleType('test')
    module.__file__ = 'test.py'
    exec(
        compile(code, module.__file__, 'exec'),
        module.__dict__,
    )  # pylint: disable=exec-used, no-member

    dynamic_ast = parser.parse('TEST')
    compiled_ast = module.TestParser().parse(
        'TEST',
        start='test',
    )  # pylint: disable=no-member
    assert dynamic_ast == compiled_ast

    dynamic_ast = parser.parse('TEST A 1')
    compiled_ast = module.TestParser().parse(
        'TEST A 1',
        start='test',
    )  # pylint: disable=no-member
    assert dynamic_ast == compiled_ast


def test_none_whitespace():
    grammar = """
        @@whitespace:: None

        start = "This is a" test;
        test = " test";
    """
    input = 'This is a test'

    parser = tatsu.compile(grammar)
    output = parser.parse(input)
    assert output == ('This is a', ' test')

    parser = generate_and_load_parser('W', grammar)
    output = parser.parse(input, parseinfo=False)
    assert output == ('This is a', ' test')


def test_sep_join():
    grammar = r"""
    @@grammar::numbers

    start
        = expression $
        ;

    expression
        = ~ ( "," )%{ digit }+
        ;

    digit = /\d+/ ;
    """
    parser = generate_and_load_parser('W', grammar)
    parser.parse('1,2,3,4', nameguard=False)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import json
from pathlib import Path
from pprint import pprint

from codegen import PostfixCodeGenerator

import tatsu
from tatsu.ast import AST
from tatsu.util import asjsons
from tatsu.walkers import NodeWalker


def simple_parse():
    grammar = Path('grammars/calc_cut.tatsu').read_text()

    parser = tatsu.compile(grammar)
    ast = parser.parse('3 + 5 * ( 10 - 20 )', trace=False, colorize=True)

    print()
    print('# SIMPLE PARSE')
    print('# AST')
    pprint(ast, width=20, indent=4)
    print()

    print('# JSON')
    print(json.dumps(ast, indent=4))
    print()


def annotated_parse():
    grammar = Path('grammars/calc_annotated.tatsu').read_text()

    parser = tatsu.compile(grammar)
    ast = parser.parse('3 + 5 * ( 10 - 20 )')

    print()
    print('# ANNOTATED AST')
    pprint(ast, width=20, indent=4)
    print()


class CalcBasicSemantics:
    def number(self, ast):
        return int(ast)

    def term(self, ast):
        if not isinstance(ast, AST):
            return ast
        elif ast.op == '*':
            return ast.left * ast.right
        elif ast.op == '/':
            return ast.left / ast.right
        else:
            raise ValueError('Unknown operator', ast.op)

    def expression(self, ast):
        if not isinstance(ast, AST):
            return ast
        elif ast.op == '+':
            return ast.left + ast.right
        elif ast.op == '-':
            return ast.left - ast.right
        else:
            raise ValueError('Unknown operator', ast.op)


def parse_with_basic_semantics():
    grammar = Path('grammars/calc_annotated.tatsu').read_text()

    parser = tatsu.compile(grammar)
    result = parser.parse('3 + 5 * ( 10 - 20 )', semantics=CalcBasicSemantics())

    print()
    print('# BASIC SEMANTICS RESULT')
    assert result == -47
    print(result)
    print()


class CalcSemantics:
    def number(self, ast):
        return int(ast)

    def addition(self, ast):
        return ast.left + ast.right

    def subtraction(self, ast):
        return ast.left - ast.right

    def multiplication(self, ast):
        return ast.left * ast.right

    def division(self, ast):
        return ast.left / ast.right


def parse_factored():
    grammar = Path('grammars/calc_factored.tatsu').read_text()

    parser = tatsu.compile(grammar)
    ast = parser.parse('3 + 5 * ( 10 - 20 )', semantics=CalcSemantics())

    print()
    print('# FACTORED SEMANTICS RESULT')
    pprint(ast, width=20, indent=4)
    print()


def parse_to_model():
    grammar = Path('grammars/calc_model.tatsu').read_text()

    parser = tatsu.compile(grammar, asmodel=True)
    model = parser.parse('3 + 5 * ( 10 - 20 )')

    print()
    print('# MODEL TYPE IS:', type(model).__name__)
    print(asjsons(model))
    print()


class CalcWalker(NodeWalker):
    def walk_object(self, node):
        return node

    def walk__add(self, node):
        return self.walk(node.left) + self.walk(node.right)

    def walk__subtract(self, node):
        return self.walk(node.left) - self.walk(node.right)

    def walk__multiply(self, node):
        return self.walk(node.left) * self.walk(node.right)

    def walk__divide(self, node):
        return self.walk(node.left) / self.walk(node.right)


def parse_and_walk_model():
    grammar = Path('grammars/calc_model.tatsu').read_text()

    parser = tatsu.compile(grammar, asmodel=True)
    model = parser.parse('3 + 5 * ( 10 - 20 )')

    print()
    print('# WALKER RESULT')
    result = CalcWalker().walk(model)
    assert result == -47, result
    print(result)
    print()


def parse_and_translate():
    grammar = Path('grammars/calc_model.tatsu').read_text()

    parser = tatsu.compile(grammar, asmodel=True)
    model = parser.parse('3 + 5 * ( 10 - 20 )')

    codegen = PostfixCodeGenerator()
    codegen.walk(model)
    postfix = codegen.printed_text()

    print()
    print('# TRANSLATED TO POSTFIX')
    print(postfix)


def main():
    simple_parse()
    annotated_parse()
    parse_with_basic_semantics()
    parse_factored()
    parse_to_model()
    parse_and_walk_model()
    parse_and_translate()


if __name__ == '__main__':
    main()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# from: https://www.unicode.org/charts/nameslist/n_2500.html
from __future__ import annotations

from tatsu.util import string


def test_visible_width():
    assert string.unicode_display_len("abc") == 3
    assert string.unicode_display_len("Python") == 6
    assert string.unicode_display_len("蛇") == 2
    assert string.unicode_display_len("🐍 Py") == 5



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import ast as stdlib_ast
from collections.abc import Callable, Iterable
from contextlib import contextmanager, suppress
from typing import Any

from .. import buffering, tokenizing
from ..ast import AST
from ..buffering import Buffer
from ..collections import BoundedDict
from ..exceptions import (
    FailedExpectingEndOfText,
    FailedLeftRecursion,
    FailedLookahead,
    FailedParse,
    FailedPattern,
    FailedSemantics,
    FailedToken,
    KeywordError,
    OptionSucceeded,
    ParseError,
    ParseException,
)
from ..infos import Alert, ParseInfo, ParserConfig, RuleInfo
from ..tokenizing import NullTokenizer, Tokenizer
from ..util import regexp, safe_name, trim
from ..util.abctools import is_list, left_assoc, prune_dict, right_assoc
from ..util.deprecate import deprecated
from ..util.safeeval import is_eval_safe, safe_builtins, safe_eval
from ..util.typing import boundcall
from ..util.undefined import Undefined
from .infos import MemoKey, RuleResult, closure
from .state import ParseState, ParseStateStack
from .tracing import EventTracer, EventTracerImpl

__all__: list[str] = ['ParseContext']


class ParseContext:
    def __init__(
        self,
        /,
        *,
        config: ParserConfig | None = None,
        **settings: Any,
    ) -> None:
        super().__init__()

        config = ParserConfig.new(config, **settings)
        if config.tokenizercls is None:
            config = config.override(tokenizercls=Buffer)
        self._config: ParserConfig = config
        self._active_config: ParserConfig = self._config

        self._tokenizer: Tokenizer = NullTokenizer()
        self._semantics: type | None = config.semantics
        self._initialize_caches()

        self._tracer: EventTracer = EventTracer()
        self.update_tracer()

    def _initialize_caches(self) -> None:
        self._states = ParseStateStack()
        self._ruleinfo_stack: list[RuleInfo] = []

        self.substate: Any = None
        self._lookahead: int = 0
        self._furthest_exception: FailedParse | None = None

        self._clear_memoization_caches()

    @property
    def config(self):
        return self._active_config

    @property
    def active_config(self) -> ParserConfig:
        return self._active_config

    @property
    def self_config(self) -> ParserConfig:
        return self._config

    @property
    def semantics(self) -> Any:
        return self._semantics

    @property
    def keywords(self) -> set[str]:
        return self._keywords

    @property
    def tracer(self):
        return self._tracer

    @property
    def tokenizer(self) -> Tokenizer:
        return self._tokenizer

    @property
    def tokenizercls(self) -> type[Tokenizer]:
        if self.config.tokenizercls is None:
            return buffering.Buffer
        else:
            return self.config.tokenizercls

    @property
    def last_node(self) -> Any:
        return self.states.last_node

    @last_node.setter
    def last_node(self, value: Any) -> None:
        self.states.last_node = value

    @property
    def pos(self) -> int:
        return self._tokenizer.pos

    @property
    def line(self) -> int:
        return self._tokenizer.line

    @property
    def states(self) -> ParseStateStack:
        return self._states

    @property
    def state(self) -> ParseState:
        return self.states.top

    @property
    def ast(self) -> AST:
        return self.state.ast

    @ast.setter
    def ast(self, value: AST) -> None:
        self.state.ast = value

    @property
    def cst(self) -> Any:
        return self.state.cst

    @cst.setter
    def cst(self, value: Any) -> None:
        self.state.cst = value

    def update_tracer(self) -> EventTracer:
        tracer = EventTracerImpl(
            self._tokenizer,
            self._ruleinfo_stack,
            config=self.config,
        )
        self._tracer = tracer
        return self.tracer

    def _reset(self, config: ParserConfig) -> ParserConfig:
        self._initialize_caches()
        self._keywords: set[str] = set(config.keywords)
        self._semantics = config.semantics
        if hasattr(self.semantics, 'set_context'):
            self.semantics.set_context(self)
        return config

    def _clear_memoization_caches(self) -> None:
        self._memos: BoundedDict[MemoKey, RuleResult | ParseException] = BoundedDict(
            self.config.memo_cache_size,
        )
        self._results: dict[MemoKey, RuleResult | ParseException] = {}

    def _set_furthest_exception(self, e: FailedParse) -> None:
        if not self._furthest_exception or e.pos > self._furthest_exception.pos:
            self._furthest_exception = e

    def parse(
        self,
        text: str | Tokenizer,
        /,
        *,
        config: ParserConfig | None = None,
        **settings: Any,
    ) -> Any:
        config = self.config.override_config(config)
        config = config.override(**settings)
        self._active_config = config
        self.update_tracer()
        try:
            self._reset(config)
            if isinstance(text, tokenizing.Tokenizer):
                tokenizer = text
            elif issubclass(config.tokenizercls, NullTokenizer):
                tokenizer = Buffer(text=text, config=config, **settings)
            elif text is not None:
                cls = self.tokenizercls
                tokenizer = cls(text, config=config, **settings)
            else:
                raise ParseError('No tokenizer or text')
            self._tokenizer = tokenizer

            if self.config.semantics and hasattr(self.config.semantics, 'set_context'):
                self.config.semantics.set_context(self)

            start: str = self.config.effective_start_rule_name() or 'start'
            rule = self._find_rule(start)
            return rule()

        except FailedParse as e:
            self._set_furthest_exception(e)
            if isinstance(self._furthest_exception, Exception):
                raise self._furthest_exception from e
            raise
        finally:
            self._clear_memoization_caches()
            self._active_config = self._config
            self.update_tracer()
            if self.config.semantics and hasattr(self.config.semantics, 'set_context'):
                self.config.semantics.set_context(None)

    def goto(self, pos: int) -> None:
        self._tokenizer.goto(pos)

    def _next(self) -> Any:
        return self._tokenizer.next()

    def next_token(self, ruleinfo: RuleInfo | None = None) -> None:
        if not (ruleinfo and ruleinfo.is_token_rule()):
            self._tokenizer.next_token()

    def _define(
        self,
        keys: Iterable[str],
        list_keys: Iterable[str] | None = None,
    ) -> None:
        # NOTE: called by generated parsers
        return self.states.define(keys, list_keys)

    def setname(self, name: str) -> None:
        # NOTE: called by generated parsers
        self.states.setname(name)

    def addname(self, name: str) -> None:
        # NOTE: called by generated parsers
        self.states.addname(name)

    # bw compatibility
    @deprecated(replacement=setname)
    def name_last_node(self, name: str):  # bw-compat
        self.setname(name)

    # bw compatibility
    @deprecated(replacement=addname)
    def add_last_node_to_name(self, name: str):  # bw-compat
        self.addname(name)

    def pushstate(self, ast: Any = None) -> None:
        self.states.push(pos=self.pos, ast=ast)

    def popstate(self) -> ParseState:
        return self.states.pop()

    def mergestate(self) -> ParseState:
        return self.states.merge(pos=self.pos)

    def undostate(self) -> None:
        self.states.pop()
        self.tokenizer.goto(self.state.pos)

    def _cut(self) -> None:
        self.states.set_cut_seen()
        self.tracer.trace_cut()

        def prune(cache: dict[Any, Any], cut_pos: int) -> None:
            prune_dict(
                cache,
                lambda k, v: k[0] < cut_pos and not isinstance(v, FailedLeftRecursion),
            )

        if self.config.prune_memos_on_cut:
            prune(self._memos, self.pos)

    def _memoization(self) -> bool:
        return self.config.memoization and (
            self.config.memoize_lookaheads or self._lookahead == 0
        )

    def _find_rule(self, name: str) -> Callable[[], Any]:
        raise NotImplementedError

    def _find_semantic_action(self, name: str) -> Callable[..., Any] | None:
        if not self.semantics:
            return None

        action = getattr(self.semantics, safe_name(name), None)
        if not callable(action):
            action = getattr(self.semantics, '_default', None)

        if not callable(action):
            action = None

        return action

    def newexcept(
        self,
        item: Any,
        exclass: type[FailedParse] = FailedParse,
    ) -> FailedParse:
        if issubclass(exclass, FailedLeftRecursion):
            rulestack: list[str] = []
        else:
            rulestack = [r.name for r in reversed(self._ruleinfo_stack)]
        return exclass(self.tokenizer.lineinfo(), rulestack, item)

    # bw compatibility
    @deprecated(replacement=newexcept)
    def _error(
        self,
        item: Any,
        exclass: type[FailedParse] = FailedParse,
    ) -> FailedParse:
        raise self.newexcept(item, exclass)

    def _fail(self):
        raise self.newexcept('fail')

    def _make_parseinfo(self, name: str, pos: int) -> ParseInfo:
        endpos = self.pos
        return ParseInfo(
            tokenizer=self.tokenizer,
            rule=name,
            pos=pos,
            endpos=endpos,
            line=self.tokenizer.posline(pos),
            endline=self.tokenizer.posline(endpos),
            alerts=self.state.alerts,
        )

    def _set_parseinfo(self, node: Any, name: str, pos: int):
        if not self.config.parseinfo:
            return
        if hasattr(node, 'set_parseinfo'):
            parseinfo = self._make_parseinfo(name, pos)
            node.set_parseinfo(parseinfo)
        elif hasattr(node, 'parseinfo'):
            parseinfo = self._make_parseinfo(name, pos)
            node.parseinfo = parseinfo

    @property
    def ruleinfo(self) -> RuleInfo:
        return self._ruleinfo_stack[-1]

    def memokey(self) -> MemoKey:
        return MemoKey(self.pos, self.ruleinfo, self.substate)

    def _memoize(
        self,
        key: MemoKey,
        memo: RuleResult | ParseException,
    ) -> RuleResult | ParseException:
        if self._memoization() and key.ruleinfo.is_memoizable:
            self._memos[key] = memo
        return memo

    def _save_result(self, key: MemoKey, result: RuleResult) -> None:
        if is_list(result.node):
            result = result._replace(node=closure(result.node))
        self._results[key] = result

    def _set_left_recursion_guard(self, key: MemoKey) -> None:
        if not self.config.left_recursion:
            return
        ex = self.newexcept(key.ruleinfo.name, exclass=FailedLeftRecursion)
        self._memoize(key, ex)

    def _call(self, ruleinfo: RuleInfo) -> Any:
        self._ruleinfo_stack += [ruleinfo]
        pos = self.pos
        try:
            self.tracer.trace_entry()
            self.last_node = None

            result = self._recursive_call(ruleinfo)

            self.goto(result.newpos)
            self.substate = result.newstate
            self.states.append(result.node)

            self.tracer.trace_success()

            return result.node
        except FailedParse as e:
            self.goto(pos)
            self._set_furthest_exception(e)
            self.tracer.trace_failure(e)
            raise
        finally:
            self._ruleinfo_stack.pop()

    def _clear_recursion_errors(self) -> None:
        def filter_func(key: MemoKey, value: Any) -> bool:
            return isinstance(value, FailedLeftRecursion)

        prune_dict(self._memos, filter_func)

    def _found_left_recursion(self, ruleinfo: RuleInfo) -> bool:
        return any(ri.name == ruleinfo.name for ri in self._ruleinfo_stack)

    def _recursive_call(self, ruleinfo: RuleInfo) -> RuleResult:
        self.next_token(ruleinfo)
        key: MemoKey = self.memokey()

        if not ruleinfo.is_leftrec:
            return self._rule_call(ruleinfo, key)
        elif not self.config.left_recursion:
            raise self.newexcept('Left recursion detected', exclass=FailedLeftRecursion)

        result: RuleResult | ParseException | None = self._results.get(key)
        if isinstance(result, RuleResult):
            return result
        elif isinstance(result, Exception):
            raise result

        result = self.newexcept(ruleinfo.name, FailedLeftRecursion)
        self._results[key] = result

        initial = self.pos
        lastpos = -1
        while True:
            self._clear_recursion_errors()
            try:
                new_result = self._rule_call(ruleinfo, key)
                self.goto(initial)
            except FailedParse:
                break

            if new_result.newpos > lastpos:
                self._save_result(key, new_result)
                lastpos = new_result.newpos
                result = new_result
            else:
                break

        if isinstance(result, Exception):
            raise result

        return result

    def _rule_call(self, ruleinfo: RuleInfo, key: MemoKey) -> RuleResult:
        result = self._memos.get(key)
        if isinstance(result, Exception):
            raise result
        if isinstance(result, RuleResult):
            return result

        self._set_left_recursion_guard(key)

        self.pushstate(ast=AST())
        try:
            self.next_token(ruleinfo)

            with self.states.cutscope():
                ruleinfo.impl(self)

            node = self.state.node
            node = self._semantics_call(ruleinfo, node)
            self._set_parseinfo(node, ruleinfo.name, key.pos)

            result = RuleResult(node, self.pos, self.substate)
            self._memoize(key, result)

            return result
        except FailedSemantics as e:
            ex = self.newexcept(str(e))
            self._memoize(key, ex)
            raise ex from e
        except ParseException as e:
            self._memoize(key, e)
            raise
        finally:
            self.undostate()

    def _semantics_call(self, ruleinfo: RuleInfo, node: Any) -> Any:
        if ruleinfo.is_name:
            self._check_name(node)

        params = ruleinfo.params or ()
        kwparams = ruleinfo.kwparams or {}
        action = self._find_semantic_action(ruleinfo.name)
        if action:
            return boundcall(action, {}, node, *params, **kwparams)
        else:
            return node

    def _token(self, token: str) -> str:
        self.next_token()
        if self.tokenizer.match(token) is None:
            self.tracer.trace_match(token, failed=True)
            raise self.newexcept(token, exclass=FailedToken)
        self.tracer.trace_match(token)
        self.states.append(token)
        return token

    def _constant(self, literal: Any) -> Any:
        self.next_token()
        self.tracer.trace_match(literal)

        if not isinstance(literal, str):
            self.states.append(literal)
            return literal
        literal = str(literal)  # for type linters

        context: dict[str, Any] = (
            safe_builtins()
            | getattr(self.semantics, 'safe_context', lambda: {})()
            | self.ast
            if isinstance(self.ast, AST)
            else {}
        )

        expression = Undefined
        result = literal
        while result != expression:
            expression = result
            if not isinstance(expression, str):
                break

            expression = trim(expression)
            with suppress(ValueError, SyntaxError):
                result = stdlib_ast.literal_eval(expression.strip())
                assert result is not Undefined
                continue

            try:
                fstr_expression = f'''f{expression!r}'''

                if is_eval_safe(fstr_expression, context):
                    result = safe_eval(fstr_expression, context)

                if result == expression and is_eval_safe(expression, context):
                    # NOTE: No f'{xyz}' evaluations occurred
                    result = safe_eval(expression, context)
            except Exception as e:
                raise FailedSemantics(
                    f'Error evaluating constant {literal!r}: {e}',
                ) from e

        self.states.append(result)
        return result

    def _alert(self, message: str, level: int) -> None:
        self.next_token()
        self.tracer.trace_match(f'{"^" * level}`{message}`', failed=True)
        self.state.alerts.append(Alert(message=message, level=level))

    def _pattern(self, pattern: str) -> Any:
        token = self.tokenizer.matchre(pattern)
        if token is None:
            self.tracer.trace_match('', pattern, failed=True)
            raise self.newexcept(f'Expecting {regexp(pattern)}', exclass=FailedPattern)
        self.tracer.trace_match(token, pattern)
        self.states.append(token)
        return token

    def eof(self) -> bool:
        return self.tokenizer.atend()

    def eol(self) -> bool:
        return self.tokenizer.ateol()

    def _check_eof(self) -> None:
        self.next_token()
        if not self.tokenizer.atend():
            raise self.newexcept(
                'Expecting end of text',
                exclass=FailedExpectingEndOfText,
            )

    @contextmanager
    def _try(self):
        s = self.substate
        self.pushstate(ast=AST(self.ast))
        self.last_node = None
        try:
            yield
            self.mergestate()
        except FailedParse:
            self.undostate()
            self.substate = s
            raise

    def _no_more_options(self) -> bool:
        """
        Used by the Python code generator so there are
        no unconditional:
            ```
            raise self.newexception(...)
            ```
        that fool the syntax highlighting of editors
        """
        return True

    @contextmanager
    def _option(self):
        self.last_node = None
        try:
            with self._try():
                yield
            raise OptionSucceeded()
        except FailedParse:
            if not self.states.cut_seen():
                pass
            else:
                raise

    @contextmanager
    def _choice(self):
        with suppress(OptionSucceeded), self.states.cutscope():
            yield

    @contextmanager
    def _optional(self):
        with self._choice(), self._option(), self.states.cutscope():
            yield

    @contextmanager
    def _group(self):
        self.pushstate()
        try:
            with self.states.cutscope():
                yield
            self.mergestate()
        except ParseException:
            self.undostate()
            raise

    @contextmanager
    def _if(self):
        s = self.substate
        self.pushstate()
        self._lookahead += 1
        try:
            yield
        finally:
            self.undostate()
            self._lookahead -= 1
            self.substate = s

    @contextmanager
    def _ifnot(self):
        try:
            with self._if():
                yield
        except ParseException:
            pass
        else:
            raise self.newexcept('', exclass=FailedLookahead)

    @contextmanager
    def _setname(self, name: str):
        yield
        self.setname(name)

    @contextmanager
    def _addname(self, name: str):
        yield
        self.addname(name)

    def _isolate(self, block: Callable[[], Any], drop: bool = False) -> Any:
        self.pushstate()
        try:
            block()
            return closure(self.cst) if is_list(self.cst) else self.cst
        finally:
            ast = self.ast
            self.popstate()
            self.ast = ast

    def _repeat(
        self,
        block: Callable[[], Any],
        prefix: Callable[[], Any] | None = None,
        dropprefix: bool = False,
    ) -> None:
        while True:
            with self._choice():
                with self._option():
                    p = self.pos

                    if prefix:
                        pcst = self._isolate(prefix, drop=dropprefix)
                        if not dropprefix:
                            self.states.append(pcst)
                        self._cut()

                    cst = self._isolate(block)
                    self.states.append(cst)

                    if self.pos == p:
                        raise self.newexcept('empty closure')
                return

    def _closure(
        self,
        block: Callable[[], Any],
        sep: Callable[[], Any] | None = None,
        omitsep: bool = False,
    ) -> Any:
        self.pushstate()
        try:
            self.cst = []
            with self.states.cutscope():
                with self._optional():
                    block()
                    self.cst = [self.cst]
                self._repeat(block, prefix=sep, dropprefix=omitsep)
            self.cst = closure(self.cst)
            return self.mergestate().cst
        except ParseException:
            self.undostate()
            raise

    def _positive_closure(
        self,
        block: Callable[[], Any],
        sep: Callable[[], Any] | None = None,
        omitsep: bool = False,
    ) -> Any:
        self.pushstate()
        try:
            with self.states.cutscope():
                block()
                self.cst = [self.cst]
                self._repeat(block, prefix=sep, dropprefix=omitsep)
            self.cst = closure(self.cst)
            return self.mergestate().cst
        except ParseException:
            self.undostate()
            raise

    def _empty_closure(self) -> closure:
        cst = closure([])
        self.states.append(cst)
        return cst

    def _gather(self, block: Callable[[], Any], sep: Callable[[], Any]) -> Any:
        return self._closure(block, sep=sep, omitsep=True)

    def _positive_gather(self, block: Callable[[], Any], sep: Callable[[], Any]) -> Any:
        return self._positive_closure(block, sep=sep, omitsep=True)

    def _join(self, block: Callable[[], Any], sep: Callable[[], Any]) -> Any:
        return self._closure(block, sep=sep)

    def _positive_join(self, block: Callable[[], Any], sep: Callable[[], Any]) -> Any:
        return self._positive_closure(block, sep=sep)

    def _left_join(self, block: Callable[[], Any], sep: Callable[[], Any]) -> Any:
        self.cst = left_assoc(self._positive_join(block, sep))
        self.last_node = self.cst
        return self.cst

    def _right_join(self, block: Callable[[], Any], sep: Callable[[], Any]) -> Any:
        self.cst = right_assoc(self._positive_join(block, sep))
        self.last_node = self.cst
        return self.cst

    def _check_name(self, name: Any) -> None:
        name_str = str(name)
        if self.config.ignorecase or self.tokenizer.ignorecase:
            name_str = name_str.upper()
        if name_str in self.keywords:
            raise self.newexcept(f'"{name_str}" is a reserved word', KeywordError)

    def _void(self) -> None:
        self.last_node = None

    def _dot(self) -> Any:
        c = self._next()
        if c is None:
            self.tracer.trace_match(c, failed=True)
            raise self.newexcept(c, exclass=FailedToken) from None
        self.tracer.trace_match(c)
        self.states.append(c)
        return c

    def _skip_to(self, block: Callable[[], Any]) -> None:
        while not self.eof():
            try:
                with self._if():
                    block()
            except FailedParse:
                pass
            else:
                break
            pos = self.pos
            self.next_token()
            if self.pos == pos:
                self._next()
        block()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

# ruff: noqa: PLW0603


class Color:
    BLACK = ''
    BLUE = ''
    CYAN = ''
    GREEN = ''
    MAGENTA = ''
    RED = ''
    RESET = ''
    WHITE = ''
    YELLOW = ''
    LIGHTBLACK_EX = ''
    LIGHTRED_EX = ''
    LIGHTGREEN_EX = ''
    LIGHTYELLOW_EX = ''
    LIGHTBLUE_EX = ''
    LIGHTMAGENTA_EX = ''
    LIGHTCYAN_EX = ''
    LIGHTWHITE_EX = ''

    BLACK_BG = ''
    BLUE_BG = ''
    CYAN_BG = ''
    GREEN_BG = ''
    MAGENTA_BG = ''
    RED_BG = ''
    RESET_BG = ''
    WHITE_BG = ''
    YELLOW_BG = ''
    LIGHTBLACK_EX_BG = ''
    LIGHTRED_EX_BG = ''
    LIGHTGREEN_EX_BG = ''
    LIGHTYELLOW_EX_BG = ''
    LIGHTBLUE_EX_BG = ''
    LIGHTMAGENTA_EX_BG = ''
    LIGHTCYAN_EX_BG = ''
    LIGHTWHITE_EX_BG = ''

    BRIGHT = ''
    DIM = ''
    NORMAL = ''
    RESET_ALL = ''


def init():
    try:
        import colorama

        colorama.init()
    except ImportError:
        return

    global Fore, Back, Style
    Fore = colorama.Fore
    Back = colorama.Back
    Style = colorama.Style

    colors = {}
    colors.update({name: value for name, value in vars(Fore).items() if name.isupper()})
    colors.update(
        {name + '_BG': value for name, value in vars(Back).items() if name.isupper()},
    )
    colors.update(
        {name: value for name, value in vars(Style).items() if name.isupper()},
    )

    for name, value in colors.items():
        setattr(Color, name, value)


class _BF:
    BLACK = ''
    BLUE = ''
    CYAN = ''
    GREEN = ''
    MAGENTA = ''
    RED = ''
    RESET = ''
    WHITE = ''
    YELLOW = ''
    LIGHTBLACK_EX = ''
    LIGHTRED_EX = ''
    LIGHTGREEN_EX = ''
    LIGHTYELLOW_EX = ''
    LIGHTBLUE_EX = ''
    LIGHTMAGENTA_EX = ''
    LIGHTCYAN_EX = ''
    LIGHTWHITE_EX = ''


class _Fore(_BF):
    pass


class _Back(_BF):
    pass


class _Style:
    BRIGHT = ''
    DIM = ''
    NORMAL = ''
    RESET_ALL = ''


Fore = _Fore
Back = _Back
Style = _Style
